﻿/* 
 * PROJECT: MMD for Java
 * --------------------------------------------------------------------------------
 * This work is based on the ARTK_MMD v0.1 
 *   PY
 * http://ppyy.hp.infoseek.co.jp/
 * py1024<at>gmail.com
 * http://www.nicovideo.jp/watch/sm7398691
 *
 * The MMD for Java is Java version MMD class library.
 * Copyright (C)2009 nyatla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * For further information please contact.
 *	http://nyatla.jp/
 *	<airmail(at)ebony.plala.or.jp>
 * 
 */
using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using jp.nyatla.nymmd.cs.types;
using jp.nyatla.nymmd.cs.core;
using jp.nyatla.nymmd.cs.struct_type;
using jp.nyatla.nymmd.cs.struct_type.pmd;

namespace jp.nyatla.nymmd.cs
{

    class DataComparator : IComparer<PmdIK>
    {
        public int Compare(PmdIK o1, PmdIK o2)
        {
            return (int)(o1.getSortVal() - o2.getSortVal());
        }
    }

    public class MmdPmdModel
    {
        private String _name;	// モデル名
        private int _number_of_vertex;	// 頂点数

        private PmdFace[] m_pFaceArray; // 表情配列
        private PmdBone[] m_pBoneArray; // ボーン配列
        private PmdIK[] m_pIKArray;    // IK配列

        private MmdVector3[] _position_array;	// 座標配列	
        private MmdVector3[] _normal_array;		// 法線配列
        private MmdTexUV[] _texture_uv;		// テクスチャ座標配列
        private PmdSkinInfo[] _skin_info_array;
        private PmdMaterial[] _materials;		// マテリアル配列

        // -----------------------------------------------------
        // コンストラクタ
        // -----------------------------------------------------
        public MmdPmdModel()
        {
            return;
        }
        public MmdPmdModel(Stream i_stream)
        {
            PmdFileData pmd = new PmdFileData(i_stream);
            initialize(pmd);
            return;
        }
        public MmdPmdModel(StreamReader i_reader)
        {
            PmdFileData pmd = new PmdFileData(i_reader);
            initialize(pmd);
            return;
        }
        public MmdPmdModel(BinaryReader i_reader)
        {
            PmdFileData pmd = new PmdFileData(i_reader);
            initialize(pmd);
            return;
        }
        public MmdPmdModel(PmdFileData pmd)
        {
            initialize(pmd);
            return;
        }

        public int getNumberOfVertex()
        {
            return this._number_of_vertex;
        }
        public PmdMaterial[] getMaterials()
        {
            return this._materials;
        }
        public MmdTexUV[] getUvArray()
        {
            return this._texture_uv;
        }
        public MmdVector3[] getPositionArray()
        {
            return this._position_array;
        }
        public MmdVector3[] getNormatArray()
        {
            return this._normal_array;
        }
        public PmdSkinInfo[] getSkinInfoArray()
        {
            return this._skin_info_array;
        }
        public PmdFace[] getFaceArray()
        {
            return this.m_pFaceArray;
        }
        public PmdBone[] getBoneArray()
        {
            return this.m_pBoneArray;
        }
        public PmdIK[] getIKArray()
        {
            return this.m_pIKArray;
        }


        public PmdBone getBoneByName(String i_name)
        {
            PmdBone[] bone_array = this.m_pBoneArray;
            for (int i = 0; i < bone_array.Length; i++)
            {
                PmdBone bone = bone_array[i];
                if (bone.getName().Equals(i_name))
                    return bone;
            }
            return null;
        }
        public PmdFace getFaceByName(String i_name)
        {
            PmdFace[] face_array = this.m_pFaceArray;
            for (int i = 0; i < face_array.Length; i++)
            {
                PmdFace face = face_array[i];
                if (face.getName().Equals(i_name))
                    return face;
            }
            return null;
        }

        public String getModelName()
        {
            return this._name;
        }

        private void initialize(PmdFileData pmd)
        {
            // -----------------------------------------------------
            // モデル名をコピー
            this._name = pmd.pmd_header.szName;

            // -----------------------------------------------------
            // 頂点数をコピー
            this._number_of_vertex = pmd.number_of_vertex;

            // 頂点配列をコピー
            this._position_array = MmdVector3.createArray(this._number_of_vertex);
            this._normal_array = MmdVector3.createArray(this._number_of_vertex);
            this._texture_uv = MmdTexUV.createArray(this._number_of_vertex);
            this._skin_info_array = new PmdSkinInfo[this._number_of_vertex];

            for (int i = 0; i < this._number_of_vertex; i++)
            {
                _position_array[i].setValue(pmd.pmd_vertex[i].vec3Pos);
                _normal_array[i].setValue(pmd.pmd_vertex[i].vec3Normal);
                _texture_uv[i].setValue(pmd.pmd_vertex[i].uvTex);

                this._skin_info_array[i] = new PmdSkinInfo();
                this._skin_info_array[i].fWeight = pmd.pmd_vertex[i].cbWeight / 100.0f;
                this._skin_info_array[i].unBoneNo[0] = pmd.pmd_vertex[i].unBoneNo[0];
                this._skin_info_array[i].unBoneNo[1] = pmd.pmd_vertex[i].unBoneNo[1];
            }

            // -----------------------------------------------------
            // マテリアル配列をコピー
            this._materials = new PmdMaterial[pmd.number_of_materials];

            int indices_ptr = 0;
            for (int i = 0; i < pmd.number_of_materials; i++)
            {
                this._materials[i] = new PmdMaterial();
                this._materials[i].toon_index = pmd.pmd_material[i].toon_index;
                this._materials[i].edge_flag = pmd.pmd_material[i].edge_flag;

                this._materials[i].indices = new short[pmd.pmd_material[i].ulNumIndices];
                System.Array.Copy(pmd.indices_array, indices_ptr, this._materials[i].indices, 0, pmd.pmd_material[i].ulNumIndices);
                indices_ptr += pmd.pmd_material[i].ulNumIndices;

                this._materials[i].col4Diffuse.setValue(pmd.pmd_material[i].col4Diffuse);

                this._materials[i].col4Specular.r = pmd.pmd_material[i].col3Specular.r;
                this._materials[i].col4Specular.g = pmd.pmd_material[i].col3Specular.g;
                this._materials[i].col4Specular.b = pmd.pmd_material[i].col3Specular.b;
                this._materials[i].col4Specular.a = 1.0f;

                this._materials[i].col4Ambient.r = pmd.pmd_material[i].col3Ambient.r;
                this._materials[i].col4Ambient.g = pmd.pmd_material[i].col3Ambient.g;
                this._materials[i].col4Ambient.b = pmd.pmd_material[i].col3Ambient.b;
                this._materials[i].col4Ambient.a = 1.0f;

                this._materials[i].fShininess = pmd.pmd_material[i].fShininess;

                this._materials[i].texture_name = pmd.pmd_material[i].szTextureFileName;
                if (this._materials[i].texture_name.Length < 1)
                {
                    this._materials[i].texture_name = null;
                }
            }

            // -----------------------------------------------------
            // Bone配列のコピー
            this.m_pBoneArray = new PmdBone[pmd.number_of_bone];
            for (int i = 0; i < pmd.number_of_bone; i++)
            {
                //ボーンの親子関係を一緒に読みだすので。
                this.m_pBoneArray[i] = new PmdBone(pmd.pmd_bone[i], this.m_pBoneArray);
            }
            for (int i = 0; i < pmd.number_of_bone; i++)
            {
                this.m_pBoneArray[i].recalcOffset();
            }
 
            // -----------------------------------------------------
            // IK配列のコピー
            this.m_pIKArray = new PmdIK[pmd.number_of_ik];
            // IK配列を作成
            if (pmd.number_of_ik > 0)
            {
                for (int i = 0; i < pmd.number_of_ik; i++)
                {
                    this.m_pIKArray[i] = new PmdIK(pmd.pmd_ik[i], this.m_pBoneArray);
                }
                System.Array.Sort<PmdIK>(this.m_pIKArray, new DataComparator());
            }
            
            // -----------------------------------------------------
            // Face配列のコピー
            if (pmd.number_of_face > 0)
            {
                this.m_pFaceArray = new PmdFace[pmd.number_of_face];
                for (int i = 0; i < pmd.number_of_face; i++)
                {
                    this.m_pFaceArray[i] = new PmdFace(pmd.pmd_face[i], this.m_pFaceArray[0]);
                }
            }
            
            return;
        }

        /*private void initialize(Figure fig)
        {
            // -----------------------------------------------------
            // モデル名
            this._name = "カス子";

            // -----------------------------------------------------
            // 頂点数を数え上げる
            int[] numVertex_tso = new int[fig.TSOList.Count]; // tsoごとの頂点数
            int[] numIndices_tso = new int[fig.TSOList.Count]; // tsoごとの頂点インデックス数
            int n = 0; // tso番号
            this._number_of_vertex = 0;
            int number_of_indices = 0;
            foreach ( TSOFile tso in fig.TSOList)
            {
                numVertex_tso[n] = 0;
                for (int i = 0; i < tso.meshes.Length; i++)
                {
                for (int j = 0; j < tso.meshes[i].sub_meshes.Length; j++)
                {
                    numVertex_tso[n] += tso.meshes[i].sub_meshes[j].vertices.Length;
                    numIndices_tso[n] += (tso.meshes[i].sub_meshes[j].vertices.Length - 1) * 3;
                }
                }
                this._number_of_vertex += numVertex_tso[n];
                number_of_indices += numIndices_tso[n];
                n++;
            }

            // 頂点配列をコピーする準備
            this._position_array = MmdVector3.createArray(this._number_of_vertex);
            this._normal_array = MmdVector3.createArray(this._number_of_vertex);
            this._texture_uv = MmdTexUV.createArray(this._number_of_vertex);
            this._skin_info_array = new PmdSkinInfo[this._number_of_vertex];
            short[] indices_array = new short[number_of_indices]; // 頂点インデックス

            // 頂点配列をコピー
            n = 0; // 頂点番号
            int n_idx = 0; // 頂点インデックス番号
            foreach ( TSOFile tso in fig.TSOList)
            {
                for (int i = 0; i < tso.meshes.Length; i++)
                {
                for (int j = 0; j < tso.meshes[i].sub_meshes.Length; j++)
                {
                for (int k = 0; k < tso.meshes[i].sub_meshes[j].vertices.Length; k++)
                {
                    _position_array[n].x = -tso.meshes[i].sub_meshes[j].vertices[k].position.X;
                    _position_array[n].y = tso.meshes[i].sub_meshes[j].vertices[k].position.Y;
                    _position_array[n].z = -tso.meshes[i].sub_meshes[j].vertices[k].position.Z;
                    _normal_array[n].x = -tso.meshes[i].sub_meshes[j].vertices[k].normal.X;
                    _normal_array[n].y = tso.meshes[i].sub_meshes[j].vertices[k].normal.Y;
                    _normal_array[n].z = -tso.meshes[i].sub_meshes[j].vertices[k].normal.Z;
                    _texture_uv[n].u = tso.meshes[i].sub_meshes[j].vertices[k].u;
                    _texture_uv[n].v = tso.meshes[i].sub_meshes[j].vertices[k].v;

                    this._skin_info_array[n] = new PmdSkinInfo();
                    this._skin_info_array[n].fWeight = tso.meshes[i].sub_meshes[j].vertices[k].skin_weights[0].weight;
                    this._skin_info_array[n].unBoneNo[0] = (int)tso.meshes[i].sub_meshes[j].vertices[k].skin_weights[0].index;
                    this._skin_info_array[n].unBoneNo[1] = (int)tso.meshes[i].sub_meshes[j].vertices[k].skin_weights[1].index;

                    // -----------------------------------------------------
                    // 頂点インデックス
                    if (k >= 2)
                    {
                        if (k % 2 != 0)
                        {
                            indices_array[n_idx++] = (short)(n - 0);
                            indices_array[n_idx++] = (short)(n - 1);
                            indices_array[n_idx++] = (short)(n - 2);
                        }
                        else
                        {
                            indices_array[n_idx++] = (short)(n - 2);
                            indices_array[n_idx++] = (short)(n - 1);
                            indices_array[n_idx++] = (short)(n - 0);
                        }
                    }

                    n++; // 頂点番号を一つ進める
                }
                }
                }
            }


            // -----------------------------------------------------
            // マテリアル数
            int number_of_materials = fig.TSOList.Count;

            // マテリアル配列をコピーする準備
            this._materials = new PmdMaterial[number_of_materials];

            // マテリアル配列をコピー
            n = 0; // マテリアル番号
            int indices_ptr = 0;
            foreach ( TSOFile tso in fig.TSOList)
            {
                this._materials[n] = new PmdMaterial();
                this._materials[n].unknown = 0;

                this._materials[n].indices = new short[numIndices_tso[n]];
                System.Array.Copy(indices_array, indices_ptr, this._materials[n].indices, 0, numIndices_tso[n]);
                indices_ptr += numIndices_tso[n];

                this._materials[n].col4Diffuse.r = 0.5f;
                this._materials[n].col4Diffuse.g = 0.5f;
                this._materials[n].col4Diffuse.b = 0.5f;
                this._materials[n].col4Diffuse.a = 1.0f;

                this._materials[n].col4Specular.r = 0.5f;
                this._materials[n].col4Specular.g = 0.5f;
                this._materials[n].col4Specular.b = 0.5f;
                this._materials[n].col4Specular.a = 1.0f;

                this._materials[n].col4Ambient.r = 0.5f;
                this._materials[n].col4Ambient.g = 0.5f;
                this._materials[n].col4Ambient.b = 0.5f;
                this._materials[n].col4Ambient.a = 1.0f;

                this._materials[n].fShininess = 0.5f;

                this._materials[n].texture_name = "";
                if (this._materials[n].texture_name.Length < 1)
                {
                    this._materials[n].texture_name = null;
                }

                n++; // マテリアル番号を一つ進める
            }

            // -----------------------------------------------------
            // ボーン名対応表
            int[] tmo_node_idx = {
                0,
                84,
                25, 83,
                //28,30,32,/*33,*///34,
                ///*35,*/36,37,38,
                ///*39,*/40,41,42,
                ///*43,*/44,45,46,
                ///*47,*/48,49,50,
                ///*51,52,*/53,54,
                //56,58,60,62,
                ///*63,*/64,65,66, 
                ///*67,*/68,69,70, 
                ///*71,*/72,73,74, 
                ///*75,*/76,77,78, 
                ///*79,80,*/81,82,
                //8,10,12,
                //16,18,20};
            /*string[] vmd_bone_idx = {
                "センター",
                "頭",
                "上半身", "首",
                "右肩", "右腕", "右ひじ", "右手首", 
                "右小指１", "右小指２", "右小指３",
                "右薬指１", "右薬指２", "右薬指３", 
                "右中指１", "右中指２", "右中指３", 
                "右人指１", "右人指２", "右人指３", 
                "右親指１", "右親指２", 
                "左肩", "左腕", "左ひじ", "左手首",
                "左小指１", "左小指２", "左小指３", 
                "左薬指１", "左薬指２", "左薬指３", 
                "左中指１", "左中指２", "左中指３", 
                "左人指１", "左人指２", "左人指３", 
                "左親指１", "左親指２",
                "左足", "左ひざ", "左足首", 
                "右足", "右ひざ", "右足首"};

            // Bone配列のコピー
            TSOFile tso0 = fig.TSOList[0];
            int number_of_bone = tso0.nodes.Length;
            this.m_pBoneArray = new PmdBone[number_of_bone];
            PMD_Bone tmp_pmd_bone;
            for (int i = 0; i < tso0.nodes.Length; i++)
            {
                tmp_pmd_bone = new PMD_Bone();

                // ボーン名 （これで識別しているため必須）
                tmp_pmd_bone.szName = tso0.nodes[i].ShortName; // とりあえずtso側の名前を継承
                // 対応表に対応するものがあれば、置き換える
                for (int bn = 0; bn < tmo_node_idx.Length; bn++)
                {
                    if (tmo_node_idx[bn] == i)
                    {
                        tmp_pmd_bone.szName = vmd_bone_idx[bn];
                        break;
                    }
                }

                // 親ボーン番号 (なければ -1)
                if (tso0.nodes[i].parent != null)
                {
                    tmp_pmd_bone.nParentNo = tso0.nodes[i].parent.ID;
                }
                else
                {
                    tmp_pmd_bone.nParentNo = -1; // (なければ -1)
                }

                // 子ボーン番号 (なければ -1)
                if (tso0.nodes[i].child_nodes.Count > 0)
                {
                    tmp_pmd_bone.nChildNo = tso0.nodes[i].child_nodes[0].ID;
                }
                else
                {
                    tmp_pmd_bone.nChildNo = -1; // (なければ -1)
                }
                tmp_pmd_bone.cbKind = 1;		// ボーンの種類
                tmp_pmd_bone.unIKTarget = -1;	// IK時のターゲットボーン
                tmp_pmd_bone.vec3Position.x = -tso0.nodes[i].Translation.X;	// モデル原点からの位置
                tmp_pmd_bone.vec3Position.y = tso0.nodes[i].Translation.Y;	// モデル原点からの位置
                tmp_pmd_bone.vec3Position.z = -tso0.nodes[i].Translation.Z;	// モデル原点からの位置

                //ボーンの親子関係を一緒に読みだすので。
                this.m_pBoneArray[i] = new PmdBone(tmp_pmd_bone, this.m_pBoneArray);
            }
            for (int i = 0; i < number_of_bone; i++)
            {
                this.m_pBoneArray[i].recalcOffset();
            }
            
            /*int number_of_bone = 0;
            foreach (TSOFile tso in fig.TSOList)
            {
                number_of_bone += tso.nodes.Length;
            }

            // Bone配列のコピー
            n = 0; // ボーン番号
            this.m_pBoneArray = new PmdBone[number_of_bone];
            PMD_Bone tmp_pmd_bone;
            foreach (TSOFile tso in fig.TSOList)
            {
                if (tso.nodes.Length > 0)
                {
                    for (int i = 0; i < tso.nodes.Length; i++)
                    {
                        tmp_pmd_bone = new PMD_Bone();

                        // ボーン名 （これで識別しているため必須）
                        tmp_pmd_bone.szName = tso.nodes[i].ShortName; // とりあえずtso側の名前を継承
                        // 対応表に対応するものがあれば、置き換える
                        for (int bn = 0; bn < tmo_node_idx.Length; bn++)
                        {
                            if (tmo_node_idx[bn] == i)
                            {
                                tmp_pmd_bone.szName = vmd_bone_idx[bn];
                                break;
                            }
                        }

                        // 親ボーン番号 (なければ -1)
                        if (tso.nodes[i].parent != null)
                        {
                            tmp_pmd_bone.nParentNo = tso.nodes[i].parent.ID;
                        }
                        else
                        {
                            tmp_pmd_bone.nParentNo = -1; // (なければ -1)
                        }

                        // 子ボーン番号 (なければ -1)
                        if (tso.nodes[i].child_nodes.Count > 0)
                        {
                            tmp_pmd_bone.nChildNo = tso.nodes[i].child_nodes[0].ID;
                        }
                        else
                        {
                            tmp_pmd_bone.nChildNo = -1; // (なければ -1)
                        }
                        tmp_pmd_bone.cbKind = 1;		// ボーンの種類
                        tmp_pmd_bone.unIKTarget = -1;	// IK時のターゲットボーン
                        tmp_pmd_bone.vec3Position.x = -tso.nodes[i].Translation.X;	// モデル原点からの位置
                        tmp_pmd_bone.vec3Position.y = tso.nodes[i].Translation.Y;	// モデル原点からの位置
                        tmp_pmd_bone.vec3Position.z = -tso.nodes[i].Translation.Z;	// モデル原点からの位置

                        //ボーンの親子関係を一緒に読みだすので。
                        this.m_pBoneArray[n] = new PmdBone(tmp_pmd_bone, this.m_pBoneArray);
            
                        n++; // ボーン番号を一つ進める
                    }
                }
            }
            for (int i = 0; i < number_of_bone; i++)
            {
                this.m_pBoneArray[i].recalcOffset();
            }*/

            // -----------------------------------------------------
            // IK配列のコピー
            //int number_of_ik = 0;
            //this.m_pIKArray = new PmdIK[number_of_ik];
            // IK配列を作成
            /*if (number_of_ik > 0)
            {
                for (int i = 0; i < number_of_ik; i++)
                {
                    this.m_pIKArray[i] = new PmdIK(pmd_ik[i], this.m_pBoneArray);
                }
                System.Array.Sort<PmdIK>(this.m_pIKArray, new DataComparator());
            }*/

            /*// -----------------------------------------------------
            // Face配列のコピー
            this.m_pFaceArray = new PmdFace[pmd.number_of_face];
            // 表情配列を作成
            if (pmd.number_of_face > 0)
            {
                for (int i = 0; i < pmd.number_of_face; i++)
                {
                    this.m_pFaceArray[i] = new PmdFace(pmd.pmd_face[i], this.m_pFaceArray[0]);
                }
            }*/

            /*return;
        }*/

    }

}
