﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Diagnostics;
using System.IO;
using System.Drawing;

using TDCG;

using NyMmdUtils;
using jp.nyatla.nymmd.cs;
using jp.nyatla.nymmd.cs.struct_type;

namespace Tso2Pmd
{
    public partial class Form1 : Form
    {
        Viewer viewer;

        string[] category = {
            "身体",
            "前髪",
            "後髪",
            "頭皮(生え際)",
            "瞳",
            "ブラ",
            "全身下着・水着",
            "パンツ",
            "靴下",
            "上衣(シャツ等)",
            "全身衣装(ナース服等)",
            "上着オプション(エプロン等)",
            "下衣(スカート等)",
            "尻尾",
            "靴",
            "頭部装備(帽子等)",
            "眼鏡",
            "首輪",
            "手首",
            "背中(羽など)",
            "アホ毛類",
            "眼帯",
            "タイツ・ガーター",
            "腕装備(手甲など)",
            "リボン",
            "手持ちの小物or背景",
            "眉毛",
            "ほくろ",
            "八重歯",
            "イヤリング類"};

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            // 出力フォルダ設定
            radioButton1.Checked = true;
            textBox_Folder.Enabled = false;
            button_Folder.Enabled = false;

            // 体型ファイル設定
            textBox_Taikei.Enabled = false;
            button_Taikei.Enabled = false;

            // Viewerインスタンス作成
            viewer = new Viewer();

            // progressBar1の設定
            progressBar1.Minimum = 1;
            progressBar1.Maximum = 10;
        }

        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Taikei.Checked == true)
            {
                textBox_Taikei.Enabled = true;
                button_Taikei.Enabled = true;
            }
            else
            {
                textBox_Taikei.Enabled = false;
                button_Taikei.Enabled = false;
            }
        }

        private void button4_Click(object sender, EventArgs e)
        {
            using (FolderBrowserDialog fbd = new FolderBrowserDialog())
            {
                //最初に選択するフォルダを指定する
                if (textBox_Folder.Text != "")
                    fbd.SelectedPath = textBox_Folder.Text;

                if (fbd.ShowDialog() != DialogResult.OK)
                {
                    return;
                }
                textBox_Folder.Text = fbd.SelectedPath;
            }
        }

        private void panel1_DragEnter(object sender, DragEventArgs e)
        {
            //コントロール内にドラッグされたとき実行される
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
                //ドラッグされたデータ形式を調べ、ファイルのときはコピーとする
                e.Effect = DragDropEffects.Copy;
            else
                //ファイル以外は受け付けない
                e.Effect = DragDropEffects.None;
            
            panel1.BackColor = Color.DarkTurquoise;
        }

        private void panel1_DragDrop(object sender, DragEventArgs e)
        {
            //ドロップされたデータがファイルか調べる
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] fileName
                    = (string[])e.Data.GetData(DataFormats.FileDrop);

                // 拡張子が".tdcgsav.png"であるか確認する
                if (System.IO.Path.GetExtension(fileName[0]) == ".png" &&
                    System.IO.Path.GetExtension(
                        System.IO.Path.GetFileNameWithoutExtension(fileName[0])) == ".tdcgsav")
                {
                    //ドロップされたヘビーセーブを読み込む
                    TransFile(fileName[0]);
                }
                else
                {
                    MessageBox.Show("拡張子が.tdcgsav.pngではありません。");
                }
            }

            panel1.BackColor = Color.LightCyan;
        }

        private void panel1_DragLeave(object sender, EventArgs e)
        {
            panel1.BackColor = Color.LightCyan;
        }

        private void TransFile(string file_name)
        {
            try
            {
                // -----------------------------------------------------
                // 進行状況を表示する
                progressBar1.Value++;
                labelProgress.Text = "ヘビーセーブファイルを読み込んでいます…。";
                this.Update();

                // ヘビーセーブファイルを開く
                viewer.ClearFigureList(); // 前回読み込んだフィギュアを消去
                List<Figure> fig_list = viewer.LoadPNGFile(file_name);
                // ボーンをTMOFileよりセットアップ
                fig_list[0].UpdateNodeMapAndBoneMatrices();

                // カテゴリを得る
                List<string> opt = new List<string>();
                PNGFile png = new PNGFile();
                png.Ftso += delegate(Stream dest, int extract_length, byte[] opt1)
                {
                    opt.Add(category[(int)opt1[0]]);
                };
                png.Load(file_name);

                // -----------------------------------------------------
                // 進行状況を表示する
                progressBar1.Value++;
                labelProgress.Text = "体型ファイルを読み込み、ボーンをセットしています…。";
                this.Update();
                
                // 特殊体型ファイルを使用する場合は、読み込んで、ボーンをセットアップ
                if (checkBox_Taikei.Checked == true)
                {
                    // 特殊体型ファイルを読み込む
                    fig_list[0].Tmo.Load(textBox_Taikei.Text);

                    // ボーンをTMOFileよりセットアップ
                    fig_list[0].UpdateNodeMapAndBoneMatrices();

                    // つま先と地面の位置の差分だけシフトして、足の裏が地面に着くようにする
                    float shift_y = fig_list[0].Tmo.FindNodeByName("W_LeftToes_End").combined_matrix.M42;
                    fig_list[0].Move(0.0f, -shift_y + 0.5f, 0.0f);
                }

                // -----------------------------------------------------
                // 進行状況を表示する
                progressBar1.Value++;
                labelProgress.Text = "表情のTmoファイル群のパスを取得しています…。";
                this.Update();
                
                // 表情ファイル郡のパスを読みとる
                List<String[]> FaceFilePaths = new List<String[]>();
                FaceFilePaths.Add( System.IO.Directory.GetFiles(Application.StartupPath + "/表情/まゆ", "*.tmo") );
                FaceFilePaths.Add( System.IO.Directory.GetFiles(Application.StartupPath + "/表情/目", "*.tmo") );
                FaceFilePaths.Add( System.IO.Directory.GetFiles(Application.StartupPath + "/表情/リップ", "*.tmo") );
                FaceFilePaths.Add( System.IO.Directory.GetFiles(Application.StartupPath + "/表情/その他", "*.tmo") );

                // -----------------------------------------------------
                // 省略する頂点グループを指定する
                ClippingCheckList checkList = new ClippingCheckList();

                if (checkBox_Clipping.Checked == true)
                {
                    //Form2クラスのインスタンスを作成する
                    Form2 f = new Form2(fig_list[0], opt, ref checkList);
                    //Form2を表示する
                    //ここではモーダルダイアログボックスとして表示する
                    //オーナーウィンドウにthisを指定する
                    f.ShowDialog(this);
                    //フォームが必要なくなったところで、Disposeを呼び出す
                    f.Dispose();
                }

                // -----------------------------------------------------
                // 進行状況を表示する
                progressBar1.Value++;
                labelProgress.Text = "ファイルを変換しています…（この処理には時間がかかります）。";
                this.Update();

                // Tso2Pmdインスタンスを作成
                Tso2Pmd t2p = new Tso2Pmd();
                t2p.Kami_flag = checkBox_Kami.Checked;
                t2p.Chichi_flag = checkBox_Chichi.Checked;
                t2p.Skirt_flag = checkBox_Skirt.Checked;
                t2p.Edge_flag_flag = checkBox_Edge.Checked;
                t2p.Merger_flag = checkBox_Merger.Checked;
                t2p.CheckList = checkList;
                t2p.Chichi_spring = 200;
                
                string em;
                // ヘッダ情報を入力
                if (( em = t2p.InputHeader(textBox_ModelName.Text, textBox_Comment.Text) ) != "")
                {
                    MessageBox.Show(em);

                    // バーを初期状態に戻しておく
                    progressBar1.Value = progressBar1.Minimum;

                    return;
                }

                // 出力フォルダを指定
                if (radioButton1.Checked == true)
                {
                    t2p.OutputFilePath = System.IO.Path.GetDirectoryName(file_name) + "/"
                        + System.IO.Path.GetFileNameWithoutExtension(
                            System.IO.Path.GetFileNameWithoutExtension(file_name));
                    System.IO.Directory.CreateDirectory(t2p.OutputFilePath);
                }
                else if (radioButton2.Checked == true)
                {
                    t2p.OutputFilePath = System.IO.Path.GetDirectoryName(file_name);
                }
                else
                {
                    t2p.OutputFilePath = textBox_Folder.Text;
                }
                // 変換
                if (( em = t2p.Figure2PmdFileData(fig_list[0], FaceFilePaths) ) != "")
                {
                    MessageBox.Show(em);

                    // バーを初期状態に戻しておく
                    progressBar1.Value = progressBar1.Minimum;

                    return;
                }

                // -----------------------------------------------------
                // 進行状況を表示する
                progressBar1.Value = progressBar1.Value++;
                labelProgress.Text = "PMDファイルを書き出しています…。";
                this.Update();

                // 出力
                StreamWriter sw = new StreamWriter(t2p.OutputFilePath + "/" + textBox_ModelName.Text + ".pmd");;
                t2p.Pmd.WritePmdFile(sw);
                sw.Close();

                // -----------------------------------------------------
                // 進行状況を表示する
                progressBar1.Value = progressBar1.Maximum;
                labelProgress.Text = "全ての作業が完了しました。";
                this.Update();

                // メッセージボックスを表示する
                MessageBox.Show("変換を完了しました。");
            }
            catch
            {
                // エラーメッセージを表示する
                switch (progressBar1.Value)
                {
                    case 2:
                        MessageBox.Show("ヘビーセーブファイルを読み込むことができませんでした。\n"
                                      + "セーブファイルに問題がある可能性があります。");
                        break;
                    case 3:
                        MessageBox.Show("特殊体型ファイルを読み込むことができませんでした。");
                        break;
                    case 4:
                        MessageBox.Show("表情のTmoファイル群のパスを取得できませんでした。\n"
                                      + "実行ファイルと同一のフォルダに表情のTmoファイル群が\n"
                                      + "存在することを確認してください。");
                        break;
                    case 5:
                        MessageBox.Show("エラーにより、変換できませんでした。");
                        break;
                    case 6:
                        MessageBox.Show("PMDファイル出力時にエラーが生じました。");
                        break;
                }
            }

            // バーを初期状態に戻しておく
            progressBar1.Value = progressBar1.Minimum;
        }

        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {
            textBox_Folder.Enabled = false;
            button_Folder.Enabled = false;
        }

        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {
            textBox_Folder.Enabled = false;
            button_Folder.Enabled = false;
        }

        private void radioButton3_CheckedChanged(object sender, EventArgs e)
        {
            textBox_Folder.Enabled = true;
            button_Folder.Enabled = true;
        }

        private void checkBox_Taikei_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Taikei.Checked == true)
            {
                textBox_Taikei.Enabled = true;
                button_Taikei.Enabled = true;
            }
            else
            {
                textBox_Taikei.Enabled = false;
                button_Taikei.Enabled = false;
            }
        }

        private void button_Taikei_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog ofd = new OpenFileDialog())
            {
                ofd.Title = "体型ファイル(.tmo)";
                if (ofd.ShowDialog() != DialogResult.OK)
                {
                    return;
                }
                textBox_Taikei.Text = ofd.FileName;
            }
        }
    }

    public class ClippingCheckList
    {
        public int group_flag;
        public List<int> checkList = new List<int>();
    }
}
