﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using TDCG;
using jp.nyatla.nymmd.cs.types;
using jp.nyatla.nymmd.cs.struct_type.pmd;

namespace Tso2Pmd
{
    class T2PMaterialList
    {
        public List<PMD_Material> material_list = new List<PMD_Material>();

        T2PTextureList tex_list = new T2PTextureList();
        List<string> toon_name_list = new List<string>();

        public T2PMaterialList(Figure fig, string path)
        {
            // テクスチャの準備
            for (int tso_num = 0; tso_num < fig.TSOList.Count; tso_num++)
            foreach (TSOTex tex in fig.TSOList[tso_num].textures)
            {
                tex_list.Add(tex, tso_num);
            }

            // テクスチャをBitmapファイルに出力
            tex_list.Save(path);
        }
        
        // TSOSubScriptより、PMD_Materialを生成する
        // (ただし、頂点インデックス数は0となっているため、後に設定する必要がある)
        public void Add(int tso_num, TSOSubScript script, bool edge_flag)
        {
            PMD_Material pmd_m = new PMD_Material();

            // スクリプトよりシェーダパラメータを取得
            Shader shader = new Shader();
            shader.Load(script.lines);

            pmd_m.col4Diffuse = new MmdColor4(0.166f, 0.166f, 0.166f, 1.0f);
            pmd_m.fShininess = 6.0f;
            pmd_m.col3Specular = new MmdColor3(0.15f, 0.15f, 0.15f);
            pmd_m.col3Ambient = new MmdColor3(38.0f, 38.0f, 38.0f);

            if (edge_flag == true)
            {
                pmd_m.edge_flag = 1;
            }
            else
            {
                pmd_m.edge_flag = 0;
            }

            pmd_m.ulNumIndices = 0;

            // colorテクスチャ
            pmd_m.szTextureFileName = tex_list.GetFileName(tso_num, shader.ColorTex);

            // toonテクスチャ
            string file_name = tex_list.GetFileName(tso_num, shader.ShadeTex);
            if (toon_name_list.IndexOf(file_name) != -1)
            {
                // toonテクスチャファイル中でのインデックス
                pmd_m.toon_index = toon_name_list.IndexOf(file_name);
            }
            else
            {
                // toonテクスチャとするテクスチャ名を記憶
                // 後にtoonテクスチャファイル名を出力するときに使用。
                toon_name_list.Add(file_name);

                // toonテクスチャファイル中でのインデックス
                pmd_m.toon_index = toon_name_list.Count - 1;
            }

            material_list.Add(pmd_m);
        }

        // 同一のマテリアルを統合する
        public void MergerMaterials()
        {
            for (int i = 0; i < material_list.Count - 1; i++)
            {
                if (EqualMaterial(material_list[i], material_list[i+1]) == 0)
                {
                    material_list[i].ulNumIndices += material_list[i+1].ulNumIndices;
                    material_list.RemoveAt(i+1);
                    i = 0;
                }
            }
        }

        // ２つのマテリアルが等しいか判定する
        public int EqualMaterial(PMD_Material m1, PMD_Material m2)
        {
            // edge_flag
            if (m1.edge_flag != m2.edge_flag) return -1;

            // colorテクスチャ
            if (m1.szTextureFileName != m2.szTextureFileName) return -1;

            // toonテクスチャ
            if (m1.toon_index != m2.toon_index) return -1;

            return 0;
        }

        // トゥーンテクスチャファイル名を得る
        public string[] GetToonFileNameList()
        {
            string[] name_list = new string[10];

            if (toon_name_list.Count <= 10)
            {
                for (int i = 0; i < toon_name_list.Count; i++)
                {
                    name_list[i] = toon_name_list[i];
                }
                for (int i = toon_name_list.Count; i < 10; i++)
                {
                    name_list[i] = "toon" + i.ToString("00") + ".bmp";
                }
            }
            else
            {
                for (int i = 0; i < 10; i++)
                {
                    name_list[i] = toon_name_list[i];
                }
            }

            return name_list;
        }
    }
}
