﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using TDCG;

using jp.nyatla.nymmd.cs.types;
using jp.nyatla.nymmd.cs.core;
using jp.nyatla.nymmd.cs.struct_type;
using jp.nyatla.nymmd.cs.struct_type.pmd;

namespace Tso2Pmd
{
    public class Tso2Pmd
    {
        PmdFileData pmd;
        PmdFileData dummy_pmd;

        bool kami_flag = true;
        bool chichi_flag = true;
        bool skirt_flag = true;
        bool edge_flag_flag = false;
        bool merger_flag = true;
        String outputFilePath;
        ClippingCheckList checkList;
        float chichi_spring = 200.0f;

        List<PMD_RigidBody> body_list = new List<PMD_RigidBody>();
        List<PMD_Joint> joint_list = new List<PMD_Joint>();

        public PmdFileData Pmd { get { return pmd; } }
        public bool Kami_flag { set { kami_flag = value; } }
        public bool Chichi_flag { set { chichi_flag = value; } }
        public bool Skirt_flag { set { skirt_flag = value; } }
        public bool Edge_flag_flag { set { edge_flag_flag = value; } }
        public bool Merger_flag { set { merger_flag = value; } }
        public string OutputFilePath { set { outputFilePath = value; } get { return outputFilePath; } }
        public ClippingCheckList CheckList { set { checkList = value; } }
        public float Chichi_spring { set { chichi_spring = value; } }

        // -----------------------------------------------------
        // コンストラクタ
        // -----------------------------------------------------
        public Tso2Pmd()
        {
            // ダミーPMDファイルを開く
            StreamReader pmds;
            pmds = new StreamReader(Application.StartupPath + "/dummy.pmd");
            dummy_pmd = new PmdFileData(pmds);
            pmds.Close();

            // 初期化
            pmd = new PmdFileData();

            return;
        }

        // -----------------------------------------------------
        // ヘッダー情報を入力
        // -----------------------------------------------------
        public string InputHeader(string name, string comment)
        { 
            pmd.pmd_header = new PMD_Header();
            pmd.pmd_header.szMagic = "Pmd";
            pmd.pmd_header.fVersion = 1.0f;

            if (name.Length > 19) return "モデル名が20文字を超えています。";
            pmd.pmd_header.szName = name;

            if (comment.Length > 255) return "コメントが256文字を超えています。";
            pmd.pmd_header.szComment = comment;

            return "";
        }

        // -----------------------------------------------------
        // FigureデータよりPmdFileデータを作成
        // -----------------------------------------------------
        public string Figure2PmdFileData(Figure fig, List<String[]> FaceFilePaths)
        {
            // -----------------------------------------------------
            // 現存pmdファイルよりコピーするボーン名一覧
            // -----------------------------------------------------
            // ボーン名対応表
            string[] tableCorBone = new string[227];
            tableCorBone[0] = "下半身";
            tableCorBone[1] = "右足";
            tableCorBone[2] = "右足";
            tableCorBone[3] = "右足";
            tableCorBone[4] = "左足";
            tableCorBone[5] = "左足";
            tableCorBone[6] = "左足";
            tableCorBone[7] = "左足"; tableCorBone[8] = "左足"; tableCorBone[9] = "左足";
            tableCorBone[10] = "左ひざ"; tableCorBone[11] = "左ひざ";
            tableCorBone[12] = "左足首"; tableCorBone[13] = "左足首";
            tableCorBone[14] = "左つま先";
            tableCorBone[15] = "右足"; tableCorBone[16] = "右足"; tableCorBone[17] = "右足";
            tableCorBone[18] = "右ひざ"; tableCorBone[19] = "右ひざ";
            tableCorBone[20] = "右足首"; tableCorBone[21] = "右足首";
            tableCorBone[22] = "右つま先";
            tableCorBone[23] = "下半身";
            tableCorBone[24] = "上半身"; tableCorBone[25] = "上半身"; tableCorBone[26] = "上半身";
            tableCorBone[27] = "右肩"; tableCorBone[28] = "右肩";
            tableCorBone[29] = "右腕"; tableCorBone[30] = "右腕"; tableCorBone[31] = "右腕";
            tableCorBone[32] = "右ひじ"; tableCorBone[33] = "右ひじ";
            tableCorBone[34] = "右手首";
            tableCorBone[35] = "右小指１";
            tableCorBone[36] = "右小指２";
            tableCorBone[37] = "右小指３";
            tableCorBone[38] = "右小指先";
            tableCorBone[39] = "右薬指１";
            tableCorBone[40] = "右薬指２";
            tableCorBone[41] = "右薬指３";
            tableCorBone[42] = "右薬指先";
            tableCorBone[43] = "右中指１";
            tableCorBone[44] = "右中指２";
            tableCorBone[45] = "右中指３";
            tableCorBone[46] = "右中指先";
            tableCorBone[47] = "右人指１";
            tableCorBone[48] = "右人指２";
            tableCorBone[49] = "右人指３";
            tableCorBone[50] = "右人指先";
            tableCorBone[51] = "右親指１"; tableCorBone[52] = "右親指１";
            tableCorBone[53] = "右親指２";
            tableCorBone[54] = "右親指先";
            tableCorBone[55] = "左肩"; tableCorBone[56] = "左肩";
            tableCorBone[57] = "左腕"; tableCorBone[58] = "左腕"; tableCorBone[59] = "左腕";
            tableCorBone[60] = "左ひじ"; tableCorBone[61] = "左ひじ";
            tableCorBone[62] = "左手首";
            tableCorBone[63] = "左小指１";
            tableCorBone[64] = "左小指２";
            tableCorBone[65] = "左小指３";
            tableCorBone[66] = "左小指先";
            tableCorBone[67] = "左薬指１";
            tableCorBone[68] = "左薬指２";
            tableCorBone[69] = "左薬指３";
            tableCorBone[70] = "左薬指先";
            tableCorBone[71] = "左中指１";
            tableCorBone[72] = "左中指２";
            tableCorBone[73] = "左中指３";
            tableCorBone[74] = "左中指先";
            tableCorBone[75] = "左人指１";
            tableCorBone[76] = "左人指２";
            tableCorBone[77] = "左人指３";
            tableCorBone[78] = "左人指先";
            tableCorBone[79] = "左親指１"; tableCorBone[80] = "左親指１";
            tableCorBone[81] = "左親指２";
            tableCorBone[82] = "左親指先";
            tableCorBone[83] = "首";
            tableCorBone[84] = "頭";
            tableCorBone[85] = "頭";// face_oya
            tableCorBone[86] = "頭";// sitakuti_oya, 
            tableCorBone[87] = "頭";// Ha_Down, 
            tableCorBone[88] = "頭";// Ha_Down_END, 
            tableCorBone[89] = "頭";// sitakuti_l_1, 
            tableCorBone[90] = "頭";// sitakuti_l_1_end, 
            tableCorBone[91] = "頭";// sitakuti_r_1, 
            tableCorBone[92] = "頭";// sitakuti_r_1_end, 
            tableCorBone[93] = "頭";// sita_01, 
            tableCorBone[94] = "頭";// sita_02, 
            tableCorBone[95] = "頭";// sita_03, 
            tableCorBone[96] = "頭";// sita_03_end, 
            tableCorBone[97] = "頭";// kutiyoko_r, 
            tableCorBone[98] = "頭";// kutiyoko_r_end, 
            tableCorBone[99] = "頭";// kutiyoko_l, 
            tableCorBone[100] = "頭";// kutiyoko_l_end, 
            tableCorBone[101] = "頭";// uekuti_oya, 
            tableCorBone[102] = "頭";// Ha_UP, 
            tableCorBone[103] = "頭";// Ha_UP_END, 
            tableCorBone[104] = "頭";// uekuti_l_1, 
            tableCorBone[105] = "頭";// uekuti_l_1_end, 
            tableCorBone[106] = "頭";// uekuti_r_1, 
            tableCorBone[107] = "頭";// uekuti_r_1_end, 
            tableCorBone[108] = "頭";// eyeline_sita_R, 
            tableCorBone[109] = "頭";// eyeline_sita_R_end, 
            tableCorBone[110] = "頭";// eyeline_sita_L, 
            tableCorBone[111] = "頭";// eyeline_sita_L_end, 
            tableCorBone[112] = "頭";// L_eyeline_oya_L, 
            tableCorBone[113] = "頭";// L_eyeline_2_L, 
            tableCorBone[114] = "頭";// L_eyeline_3_L, 
            tableCorBone[115] = "頭";// L_eyeline_3_L_end, 
            tableCorBone[116] = "頭";// L_eyeline_2_R, 
            tableCorBone[117] = "頭";// L_eyeline_3_R, 
            tableCorBone[118] = "頭";// L_eyeline_3_R_end, 
            tableCorBone[119] = "頭";// R_eyeline_oya_R, 
            tableCorBone[120] = "頭";// R_eyeline_2_L, 
            tableCorBone[121] = "頭";// R_eyeline_3_L, 
            tableCorBone[122] = "頭";// R_eyeline_3_L_end, 
            tableCorBone[123] = "頭";// R_eyeline_2_R, 
            tableCorBone[124] = "頭";// R_eyeline_3_R, 
            tableCorBone[125] = "頭";// R_eyeline_3_R_end, 
            tableCorBone[126] = "頭";// hana, 
            tableCorBone[127] = "頭";// hana_END, 
            tableCorBone[128] = "頭";// mayu_1_R, 
            tableCorBone[129] = "頭";// mayu_2_R, 
            tableCorBone[130] = "頭";// mayu_3_R, 
            tableCorBone[131] = "頭";// mayu_3_R_end, 
            tableCorBone[132] = "頭";// mayu_1_L, 
            tableCorBone[133] = "頭";// mayu_2_L, 
            tableCorBone[134] = "頭";// mayu_3_L, 
            tableCorBone[135] = "頭";// mayu_3_L_end, 
            tableCorBone[136] = "左目";// Me_Left_Futi, 
            tableCorBone[137] = "左目";// Me_Left, 
            tableCorBone[138] = "左目";// Me_Left_End, 
            tableCorBone[139] = "左目";// Me_Left_HighLight_A, 
            tableCorBone[140] = "左目";// Me_Left_HighLight_A_END, 
            tableCorBone[141] = "左目";// Me_Left_HighLight_B, 
            tableCorBone[142] = "左目";// Me_Left_HighLight_B_END, 
            tableCorBone[143] = "右目";// Me_Right_Futi, 
            tableCorBone[144] = "右目";// Me_Right, 
            tableCorBone[145] = "右目";// Me_Right_End, 
            tableCorBone[146] = "右目";// Me_Left_RighLight_A, 
            tableCorBone[147] = "右目";// Me_Left_RighLight_A_END, 
            tableCorBone[148] = "右目";// Me_Left_RighLight_B, 
            tableCorBone[149] = "右目";// Me_Left_RighLight_B_END, 
            tableCorBone[150] = "頭";// Kami_Oya, 
            tableCorBone[220] = "下半身";// Mata, 
            tableCorBone[221] = "下半身";// W_RightManko, 
            tableCorBone[222] = "下半身";// W_RightManko_End, 
            tableCorBone[223] = "下半身";// W_LeftManko, 
            tableCorBone[224] = "下半身";// W_LeftManko_End, 
            tableCorBone[225] = "下半身";// Anal, 
            tableCorBone[226] = "下半身";// Anal_End

            // -----------------------------------------------------
            // 現存pmdファイルよりコピーするボーン名一覧
            // 注：順番が重要で、必ず親は子の前にいなければならない！
            string[] pmd_copy_bone = {
                "センター", "センター先",              
                "下半身", "下半身先", 
                "上半身", "首", "頭", "頭先",
                "左肩", "左腕", "左ひじ", "左手首", "左手先", 
                "左小指１", "左小指２", "左小指３", "左小指先",
                "左薬指１", "左薬指２", "左薬指３", "左薬指先", 
                "左中指１", "左中指２", "左中指３", "左中指先", 
                "左人指１", "左人指２", "左人指３", "左人差指先", 
                "左親指１", "左親指２", "左親指先",
                "右肩", "右腕", "右ひじ", "右手首", "右手先",
                "右小指１", "右小指２", "右小指３", "右小指先",
                "右薬指１", "右薬指２", "右薬指３", "右薬指先",
                "右中指１", "右中指２", "右中指３", "右中指先",
                "右人指１", "右人指２", "右人指３", "右人差指先",
                "右親指１", "右親指２", "右親指先",
                "左足", "左ひざ", "左足首", "左つま先",
                "右足", "右ひざ", "右足首", "右つま先",
                "左目", "左目先",
                "右目", "右目先",
                "両目", "両目先",
                "左足ＩＫ", "左足ＩＫ先",
                "右足ＩＫ", "右足ＩＫ先", 
                "左つま先ＩＫ", "左つま先ＩＫ先",
                "右つま先ＩＫ", "右つま先ＩＫ先"};

            // -----------------------------------------------------
            // ボーンの位置合わせのための対応表
            string[] vmd_bone_idx = {
                //"センター",               
                "下半身", "下半身先", 
                "上半身", "首", "頭", "頭先",
                "左肩", "左腕", "左ひじ", "左手首", "左手先", 
                "左小指１", "左小指２", "左小指３", "左小指先",
                "左薬指１", "左薬指２", "左薬指３", "左薬指先", 
                "左中指１", "左中指２", "左中指３", "左中指先", 
                "左人指１", "左人指２", "左人指３", "左人差指先", 
                "左親指１", "左親指２", "左親指先",
                "右肩", "右腕", "右ひじ", "右手首", "右手先",
                "右小指１", "右小指２", "右小指３", "右小指先",
                "右薬指１", "右薬指２", "右薬指３", "右薬指先",
                "右中指１", "右中指２", "右中指３", "右中指先",
                "右人指１", "右人指２", "右人指３", "右人差指先",
                "右親指１", "右親指２", "右親指先",
                "左足", "左ひざ", "左足首", "左つま先",
                "右足", "右ひざ", "右足首", "右つま先", 
                "左目", "左目先",
                "右目", "右目先",
                "左足ＩＫ",
                "右足ＩＫ",
                "左つま先ＩＫ",
                "右つま先ＩＫ"};
            int[] tmo_node_idx = {
                //0,                
                24, 0,
                24,/*25,26,*/ 83, 84, 85,/*face_oya*/
                56,58,60,/*61,*/62, 67,
                63,64,65,66, 
                67,68,69,70, 
                71,72,73,74, 
                75,76,77,78, 
                /*79,*/80,81,82,
                28,30,32,/*33,*/34, 43,
                35,36,37,38,
                39,40,41,42,
                43,44,45,46,
                47,48,49,50,
                /*51,*/52,53,54,
                 8,10,12, 14/*W_LeftToes_End*/,
                16,18,20, 22/*W_RightToes_End*/, 
                137/*Me_Left*/,138/*Me_Left_End*/, 
                144/*Me_Right*/,145/*Me_Right_End*/,
                12, 
                20, 
                14/*W_LeftToes_End*/,
                22/*W_RightToes_End*/};


            // -----------------------------------------------------
            // 予め、情報をコピーするmeshを選定し、並び替えておく
            // -----------------------------------------------------
            List<TSOMesh> mesh_list = new List<TSOMesh>();
            T2PMaterialList material_list = new T2PMaterialList(fig, outputFilePath);

            int checkList_index = 0;
            int tso_num = 0;
            foreach (TSOFile tso in fig.TSOList)
            {
                switch (checkList.group_flag)
                {
                    case 0:
                        for (int script_num = 0; script_num < tso.sub_scripts.Length; script_num++)
                        {
                            if (!(checkList.checkList.IndexOf(checkList_index++) >= 0)) // チェックを入れたマテリアルは無視する
                            {
                                foreach (TSOFrame frame in tso.frames)
                                    foreach (TSOMesh mesh in frame.meshes)
                                    {
                                        if (mesh.spec == script_num)
                                        {
                                            mesh_list.Add(mesh);
                                            material_list.Add(tso_num, tso.sub_scripts[script_num], edge_flag_flag);
                                        }
                                    }
                            }
                        }
                        break;

                    case 1:
                        for (int script_num = 0; script_num < tso.sub_scripts.Length; script_num++)
                        {
                            int tmp_checkList_index = checkList_index;
                            foreach (TSOFrame frame in tso.frames)
                            {
                                if (!(checkList.checkList.IndexOf(checkList_index++) >= 0)) // チェックを入れたマテリアルは無視する
                                {
                                    foreach (TSOMesh mesh in frame.meshes)
                                    {
                                        if (mesh.spec == script_num)
                                        {
                                            mesh_list.Add(mesh);
                                            material_list.Add(tso_num, tso.sub_scripts[script_num], edge_flag_flag);
                                        }
                                    }
                                }
                            }
                            if (script_num < tso.sub_scripts.Length - 1)
                                checkList_index = tmp_checkList_index;
                        }
                        break;

                    case 2:
                        for (int script_num = 0; script_num < tso.sub_scripts.Length; script_num++)
                        {
                            int tmp_checkList_index = checkList_index;
                            foreach (TSOFrame frame in tso.frames)
                            {
                                foreach (TSOMesh mesh in frame.meshes)
                                {
                                    if (!(checkList.checkList.IndexOf(checkList_index++) >= 0)) // チェックを入れたマテリアルは無視する
                                    {
                                        if (mesh.spec == script_num)
                                        {
                                            mesh_list.Add(mesh);
                                            material_list.Add(tso_num, tso.sub_scripts[script_num], edge_flag_flag);
                                        }
                                    }
                                }
                            }
                            if (script_num < tso.sub_scripts.Length - 1)
                                checkList_index = tmp_checkList_index;
                        }
                        break;
                }
                tso_num++;
            }


            // -----------------------------------------------------
            // 表情情報（表情情報では、頂点の情報が必要になるので、
            // 頂点についていろいろやる前に初期化をやっておかねばならない）
            // -----------------------------------------------------
            // 表情に関連するBoneの最小、及び最大
            const int FACE_BONE_MIN = 86;
            const int FACE_BONE_MAX = 135;
            // 表情に関連するBone名のリスト
            List<String>[] face_bone_list = new List<String>[4];
            // 表情(1：まゆ)に関連するBone名のリスト
            face_bone_list[0] = new List<String>();
            for (int node_num = 128; node_num <= 135; node_num++)
                face_bone_list[0].Add(fig.Tmo.nodes[node_num].Name);
            // 表情(2：目)に関連するBone名のリスト
            face_bone_list[1] = new List<String>();
            for (int node_num = 108; node_num <= 125; node_num++)
                face_bone_list[1].Add(fig.Tmo.nodes[node_num].Name);
            // 表情(3：リップ)に関連するBone名のリスト
            face_bone_list[2] = new List<String>();
            for (int node_num = 86; node_num <= 107; node_num++)
                face_bone_list[2].Add(fig.Tmo.nodes[node_num].Name);
            // 表情に関連する全てのBone名のリスト(4：その他)
            face_bone_list[3] = new List<String>();
            for (int node_num = FACE_BONE_MIN; node_num <= FACE_BONE_MAX; node_num++)
                face_bone_list[3].Add(fig.Tmo.nodes[node_num].Name);

            // 表情数
            pmd.number_of_face = 1 + FaceFilePaths[0].Length + FaceFilePaths[1].Length  
                + FaceFilePaths[2].Length + FaceFilePaths[3].Length;
            pmd.pmd_face = new PMD_FACE[pmd.number_of_face];

            // 表情に関連するboneに影響を受ける頂点を数え上げる
            int numFaceVertices = 0;
            foreach (TSOMesh mesh in mesh_list)
            {
                foreach (Vertex vertex in mesh.vertices)
                foreach (SkinWeight skin_w in vertex.skin_weights)
                {
                    // 表情に関連するboneに影響を受ける頂点であれば、表情用の頂点とする
                    if (FACE_BONE_MIN <= mesh.bone_indices[skin_w.bone_index]
                        && mesh.bone_indices[skin_w.bone_index] <= FACE_BONE_MAX)
                    {
                        numFaceVertices++;
                        break;
                    }
                }
            }

            // 表情の頂点情報を代入する準備（初期化）
            TMOFile[] face_tmo = new TMOFile[pmd.number_of_face - 1];
            int n_face = 0; // 通し番号

            // baseの表情
            pmd.pmd_face[n_face++] = new PMD_FACE(numFaceVertices);

            // base以外の表情
            for (int type = 1; type <= 4; type++)
            {
                foreach (string path in FaceFilePaths[type - 1])
                {
                    pmd.pmd_face[n_face] = new PMD_FACE(numFaceVertices);
                    pmd.pmd_face[n_face].szName = System.IO.Path.GetFileNameWithoutExtension(path); // 表情名 (0x00 終端，余白は 0xFD)
                    pmd.pmd_face[n_face].cbType = type; // 分類 (0：base、1：まゆ、2：目、3：リップ、4：その他)

                    // 表情のTMOファイルを読み込んでおく
                    face_tmo[n_face - 1] = new TMOFile();
                    face_tmo[n_face - 1].Load(path);
                    face_tmo[n_face - 1].LoadTransformationMatrixFromFrame(0);

                    n_face++;
                }
            }

            // -----------------------------------------------------
            // PMD初期状態に合わせて、TMO初期状態を変更
            // -----------------------------------------------------
            // 28, W_RightShoulder, 
            //fig.Tmo.nodes[28].Rotation
            //    = Quaternion.RotationAxis( new Vector3(0.0f, 0.0f, 1.0f), (float)((10.0 / 180.0) * Math.PI));
            // 30, W_RightArm,
            fig.Tmo.nodes[30].Rotation
                = Quaternion.RotationAxis( new Vector3(0.0f, 0.0f, 1.0f), (float)((45.0 / 180.0) * Math.PI));
            // 56, W_LeftShoulder, 
            //fig.Tmo.nodes[56].Rotation
            //    = Quaternion.RotationAxis( new Vector3(0.0f, 0.0f, 1.0f), (float)(-(10.0 / 180.0) * Math.PI));
            // 58, W_LeftArm, 
            fig.Tmo.nodes[58].Rotation
                = Quaternion.RotationAxis( new Vector3(0.0f, 0.0f, 1.0f), (float)(-(45.0 / 180.0) * Math.PI));

            fig.UpdateBoneMatricesWithoutTMOFrame();

            // -----------------------------------------------------
            // 頂点＆表情デフォ＆マテリアル
            // -----------------------------------------------------
            List<PMD_Vertex> vertex_list = new List<PMD_Vertex>();
            List<short> indices = new List<short>(); // インデックスリスト

            // -----------------------------------------------------
            // 情報をコピー
            int n_mesh = 0;
            n_face = 0; // 表情の頂点の番号（通し番号）
            int prevNumIndices = 0;
            foreach (TSOMesh mesh in mesh_list)
            {
                int n_inMesh = -1; // mesh中のvertexの番号（処理の前に++するために、初期値は0でなく-1としている)
                Vector3 a = new Vector3();
                Vector3 b = new Vector3();
                Vector3 c = new Vector3();

                // Tmo中のBoneに従って、Mesh中の頂点位置及び法線ベクトルを置き換えて、書き出す
                Vertex[] trans_v = TsoUtils.TransVerticesWithTmo(mesh, fig.Tmo);

                foreach (Vertex vertex in mesh.vertices)
                {
                    // -----------------------------------------------------
                    n_inMesh++; // mesh中のvertexの番号を一つ増やす

                    // -----------------------------------------------------
                    // 頂点情報をコピー
                    PMD_Vertex pmd_v = new PMD_Vertex();

                    pmd_v.vec3Pos = PmdUtils.CopyPos(trans_v[n_inMesh].position);
                    pmd_v.vec3Normal = PmdUtils.CopyPos(trans_v[n_inMesh].normal);
                    pmd_v.uvTex.u = vertex.u;
                    pmd_v.uvTex.v = vertex.v;

                    pmd_v.cbEdge = 0;

                    float w0 = vertex.skin_weights[0].weight;
                    float w1 = vertex.skin_weights[1].weight;
                    pmd_v.cbWeight = (int)(w0 * 100 / (w0 + w1));

                    // -----------------------------------------------------
                    // スキニング
                    // bone_index_LUT[skin_weights[0].index] -> unBoneNo[0]
                    int ver_bone_num1 = (int)mesh.bone_indices[(int)vertex.skin_weights[0].bone_index];
                    if (151 <= ver_bone_num1 && ver_bone_num1 <= 219)
                    {
                        pmd_v.unBoneNo[0] = pmd_copy_bone.Length + (ver_bone_num1 - 150);
                    }
                    else
                    {
                        for (int l = 0; l < pmd_copy_bone.Length; l++)
                        {
                            if (pmd_copy_bone[l] == tableCorBone[ver_bone_num1]) pmd_v.unBoneNo[0] = l;
                        }
                    }

                    // bone_index_LUT[skin_weights[1].index] -> unBoneNo[1]
                    int ver_bone_num2 = (int)mesh.bone_indices[(int)vertex.skin_weights[1].bone_index];
                    if (151 <= ver_bone_num2 && ver_bone_num2 <= 219)
                    {
                        pmd_v.unBoneNo[1] = pmd_copy_bone.Length + (ver_bone_num2 - 150);
                    }
                    else
                    {
                        for (int l = 0; l < pmd_copy_bone.Length; l++)
                        {
                            if (pmd_copy_bone[l] == tableCorBone[ver_bone_num2]) pmd_v.unBoneNo[1] = l;
                        }
                    }

                    // -----------------------------------------------------
                    // 頂点リストに頂点を追加
                    vertex_list.Add(pmd_v);

                    // -----------------------------------------------------
                    // 頂点インデックス
                    // 頂点重複を判定し、重複している場合は以下の処理を省略する
                    a = new Vector3(b.X, b.Y, b.Z);
                    b = new Vector3(c.X, c.Y, c.Z);
                    c = new Vector3(pmd_v.vec3Pos.x, pmd_v.vec3Pos.y, pmd_v.vec3Pos.z);

                    if ((n_inMesh >= 2) &&
                        !((a.X == b.X && a.Y == b.Y && a.Z == b.Z) ||
                          (b.X == c.X && b.Y == c.Y && b.Z == c.Z) ||
                          (c.X == a.X && c.Y == a.Y && c.Z == a.Z)))
                    {
                        int i = vertex_list.Count - 1;

                        if (n_inMesh % 2 == 0)
                        {
                            indices.Add((short)(i - 0));
                            indices.Add((short)(i - 1));
                            indices.Add((short)(i - 2));
                        }
                        else
                        {
                            indices.Add((short)(i - 2));
                            indices.Add((short)(i - 1));
                            indices.Add((short)(i - 0));
                        }
                    }

                    // -----------------------------------------------------
                    // 表情情報
                    // 表情に関連するboneに影響を受ける頂点であれば、情報を記憶する
                    foreach (SkinWeight skin_w in vertex.skin_weights)
                    {
                        // 表情に関連するboneに影響を受ける頂点であれば、表情用の頂点とする
                        if (FACE_BONE_MIN <= mesh.bone_indices[skin_w.bone_index]
                            && mesh.bone_indices[skin_w.bone_index] <= FACE_BONE_MAX)
                        {
                            // 表情の頂点情報（base）
                            pmd.pmd_face[0].pVertices[n_face].ulIndex = vertex_list.Count - 1; // 表情用の頂点の番号(頂点リストにある番号)
                            pmd.pmd_face[0].pVertices[n_face].vec3Pos = PmdUtils.CopyPos(trans_v[n_inMesh].position);
                            n_face++;
                            break;
                        }
                    }
                }

                // meshごとのインデックス数を記録
                material_list.material_list[n_mesh++].ulNumIndices = indices.Count - prevNumIndices;
                prevNumIndices = indices.Count;
            }
            // -----------------------------------------------------
            // リストを配列に代入し直す
            // 頂点情報
            pmd.number_of_vertex = vertex_list.Count;
            pmd.pmd_vertex = (PMD_Vertex[])vertex_list.ToArray();
            // 頂点インデックス
            pmd.number_of_indices = indices.Count;
            pmd.indices_array = (short[])indices.ToArray();
            // マテリアル
            if (merger_flag == true) material_list.MergerMaterials();
            pmd.number_of_materials = material_list.material_list.Count;
            pmd.pmd_material = (PMD_Material[])material_list.material_list.ToArray();
            // Toonテクスチャファイル名
            pmd.toon_file_name = material_list.GetToonFileNameList();

            // 頂点数が上限を超えてないかチェックし、超えていたらエラーを出して終了
            if (pmd.number_of_vertex > 65535)
                return "頂点数(" + pmd.number_of_vertex.ToString() + ")が上限(65535)を超えています。";

                       
            // -----------------------------------------------------
            // ボーン情報
            // -----------------------------------------------------
            // Bone数取得
            pmd.number_of_bone = pmd_copy_bone.Length + (219 - 151 + 1);
            pmd.pmd_bone = new PMD_Bone[pmd.number_of_bone];

            // Bone配列取得
            TSOFile tso0 = fig.TSOList[0];

            int n = 0; // 通し番号 
            for (int i = 0; i < pmd_copy_bone.Length; i++)
            {
                foreach (PMD_Bone bone in dummy_pmd.pmd_bone)
                {
                    if (bone.szName == pmd_copy_bone[i])
                    {
                        pmd.pmd_bone[n] = new PMD_Bone();

                        pmd.pmd_bone[n].szName = bone.szName;
                        pmd.pmd_bone[n].cbKind = bone.cbKind; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動

                        // 現存pmdファイルから欲しいボーンのみ抜き出したため、
                        // 番号を付け替えなければならない。
                        if (bone.nParentNo == -1) pmd.pmd_bone[n].nParentNo = -1;
                        else pmd.pmd_bone[n].nParentNo = Array.IndexOf(pmd_copy_bone, dummy_pmd.pmd_bone[bone.nParentNo].szName); // 親ボーン番号(ない場合は-1)
                        if (bone.nChildNo == -1) pmd.pmd_bone[n].nChildNo = -1;
                        else pmd.pmd_bone[n].nChildNo = Array.IndexOf(pmd_copy_bone, dummy_pmd.pmd_bone[bone.nChildNo].szName); // 子ボーン番号(ない場合は-1)
                        if (bone.unIKTarget == 0) pmd.pmd_bone[n].unIKTarget = 0;
                        else pmd.pmd_bone[n].unIKTarget = Array.IndexOf(pmd_copy_bone, dummy_pmd.pmd_bone[bone.unIKTarget].szName);  // IKボーン番号(影響IKボーン)

                        pmd.pmd_bone[n].vec3Position.x = bone.vec3Position.x;	// モデル原点からの位置
                        pmd.pmd_bone[n].vec3Position.y = bone.vec3Position.y;	// モデル原点からの位置
                        pmd.pmd_bone[n].vec3Position.z = bone.vec3Position.z;	// モデル原点からの位置

                        for (int k = 0; k < vmd_bone_idx.Length; k++)
                        {
                            if (pmd.pmd_bone[n].szName == vmd_bone_idx[k])
                            {
                                pmd.pmd_bone[n].vec3Position
                                    = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tmo_node_idx[k]].combined_matrix); // モデル原点からの位置
                                break;
                            }
                        }

                        n++; break;
                    }
                }
            }

            // IK先ボーンの位置調整
            pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左足ＩＫ先")].vec3Position
                = new MmdVector3(
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左足ＩＫ")].vec3Position.x,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左足ＩＫ")].vec3Position.y,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左足ＩＫ")].vec3Position.z + 2.0f);
            pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右足ＩＫ先")].vec3Position
                = new MmdVector3(
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右足ＩＫ")].vec3Position.x,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右足ＩＫ")].vec3Position.y,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右足ＩＫ")].vec3Position.z + 2.0f);
            pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左つま先ＩＫ先")].vec3Position
                = new MmdVector3(
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左つま先ＩＫ")].vec3Position.x,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左つま先ＩＫ")].vec3Position.y - 2.0f,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "左つま先ＩＫ")].vec3Position.z);
            pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右つま先ＩＫ先")].vec3Position
                = new MmdVector3(
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右つま先ＩＫ")].vec3Position.x,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右つま先ＩＫ")].vec3Position.y - 2.0f,
                    pmd.pmd_bone[PmdUtils.GetBoneIDFromName(pmd, "右つま先ＩＫ")].vec3Position.z);

            string[] phys_bone_name = {
                "髪後中１", "髪後中２", "髪後中３", "髪後中４", "髪後中先", 
                "髪横左１", "髪横左２", "髪横左３", "髪横左４", "髪横左先", 
                "髪横右１", "髪横右２", "髪横右３", "髪横右４", "髪横右先", 
                "髪後左１", "髪後左２", "髪後左３", "髪後左４", "髪後左先", 
                "髪後右１", "髪後右２", "髪後右３", "髪後右４", "髪後右先", 
                "髪前左１", "髪前左２", "髪前左３", "髪前左先", 
                "髪前右１", "髪前右２", "髪前右３", "髪前右先", 
                "乳右１", "乳右２", "乳右３", "乳右４", "乳右５", "乳右先", 
                "乳左１", "乳左２", "乳左３", "乳左４", "乳左５", "乳左先", 
                "ス後左１", "ス後左２", "ス後左３", "ス後左先", 
                "ス後右１", "ス後右２", "ス後右３", "ス後右先", 
                "ス横右１", "ス横右２", "ス横右３", "ス横右先", 
                "ス前右１", "ス前右２", "ス前右３", "ス前右先", 
                "ス前左１", "ス前左２", "ス前左３", "ス前左先", 
                "ス横左１", "ス横左２", "ス横左３", "ス横左先"};

            int tso_node_num = 151;
            int phys_bone_name_idx = 0;
            // (151, kami_Back_Mid1) - (155, kami_Back_Mid4_End)
            // (156, kami_Front_L1) - (160, kami_Front_L4_End)
            // (161, kami_Front_R1) - (165, kami_Front_R4_End)
            // (166, kami_Back_L1) - (170, kami_Back_L4_End)
            // (171, kami_Back_R1) - (175, kami_Back_R4_End)
            for (int i = 0; i < 5; i++)
            {
                // 根元
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = PmdUtils.GetBoneIDFromName(pmd, "頭"); // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;

                // 中間
                for (int j = 0; j < 3; j++)
                {
                    pmd.pmd_bone[n] = new PMD_Bone();
                    pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                    pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                    pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                    pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                    pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                    pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                    n++; tso_node_num++;
                }

                // 先端
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = 0; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 7; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;
            }

            // (176, kami_Front_Mid1_L) - (179, kami_Front_Mid3_End_L)
            // (180, kami_Front_Mid1_R) - (183, kami_Front_Mid3_End_R)
            for (int i = 0; i < 2; i++)
            {
                // 根元
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = PmdUtils.GetBoneIDFromName(pmd, "頭"); // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;
                
                // 中間
                for (int j = 0; j < 2; j++)
                {
                    pmd.pmd_bone[n] = new PMD_Bone();
                    pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                    pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                    pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                    pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                    pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                    pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                    n++; tso_node_num++;
                }

                // 先端
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = 0; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 7; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;
            }
            
            // (184, Chichi_Right1) - (189, Chichi_Right5_end)
            // (190, Chichi_Left1) - (195, Chichi_Left5_End)
            for (int i = 0; i < 2; i++)
            {
                // 根元
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = PmdUtils.GetBoneIDFromName(pmd, "上半身"); // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;
                
                // 中間
                for (int j = 0; j < 4; j++)
                {
                    pmd.pmd_bone[n] = new PMD_Bone();
                    pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                    pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                    pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                    pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                    pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                    pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                    n++; tso_node_num++;
                }

                // 先端
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = 0; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 7; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;
            }
            
            // (196, skirt_LeftB01) - (199, skirt_LeftB03_end)
            // (200, skirt_RightB01) - (203, skirt_RightB03_end)
            // (204, skirt_RightS01) - (207, skirt_RightS03_end)
            // (208, skirt_RightF01) - (211, skirt_RightF03_end)
            // (212, skirt_LeftF01) - (215, skirt_LeftF03_end)
            // (216, skirt_LeftS01) - (219, skirt_LeftS03_end)
            for (int i = 0; i < 6; i++)
            {
                // 根元
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = PmdUtils.GetBoneIDFromName(pmd, "下半身"); // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;
                
                // 中間
                for (int j = 0; j < 2; j++)
                {
                    pmd.pmd_bone[n] = new PMD_Bone();
                    pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                    pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                    pmd.pmd_bone[n].nChildNo = n+1; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                    pmd.pmd_bone[n].cbKind = 0; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                    pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                    pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                    n++; tso_node_num++;
                }

                // 先端
                pmd.pmd_bone[n] = new PMD_Bone();
                pmd.pmd_bone[n].szName = phys_bone_name[phys_bone_name_idx++];
                pmd.pmd_bone[n].nParentNo = n-1; // 親ボーン番号(ない場合は0xFFFF)
                pmd.pmd_bone[n].nChildNo = 0; // tail位置のボーン番号(チェーン末端の場合は0xFFFF) // 親：子は1：多なので、主に位置決め用
                pmd.pmd_bone[n].cbKind = 7; // ボーンの種類 0:回転 1:回転と移動 2:IK 3:不明 4:IK影響下 5:回転影響下 6:IK接続先 7:非表示 8:捻り 9:回転運動
                pmd.pmd_bone[n].unIKTarget = 0; // IKボーン番号(影響IKボーン。ない場合は0)
                pmd.pmd_bone[n].vec3Position = PmdUtils.CopyMat2Pos(fig.Tmo.nodes[tso_node_num].combined_matrix);	// モデル原点からの位置
                n++; tso_node_num++;
            }
                
            // -----------------------------------------------------
            // IK配列
            // -----------------------------------------------------
            pmd.number_of_ik = 4;

            pmd.pmd_ik = new PMD_IK[pmd.number_of_ik];

            n = 0; // 通し番号
            pmd.pmd_ik[n] = new PMD_IK();
            pmd.pmd_ik[n].nTargetNo = Array.IndexOf(pmd_copy_bone, "右足ＩＫ");	// IKボーン番号
            pmd.pmd_ik[n].nEffNo = Array.IndexOf(pmd_copy_bone, "右足首");		// IKターゲットボーン番号 // IKボーンが最初に接続するボーン
            pmd.pmd_ik[n].cbNumLink = 2;	// IKチェーンの長さ(子の数)
            pmd.pmd_ik[n].unCount = 40;      // 再帰演算回数 // IK値1
            pmd.pmd_ik[n].fFact = 0.5f;       // IKの影響度 // IK値2
            int[] tem_array1 = { Array.IndexOf(pmd_copy_bone, "右ひざ"), Array.IndexOf(pmd_copy_bone, "右足") }; // IK影響下のボーン番号
            pmd.pmd_ik[n++].punLinkNo = tem_array1;

            pmd.pmd_ik[n] = new PMD_IK();
            pmd.pmd_ik[n].nTargetNo = Array.IndexOf(pmd_copy_bone, "左足ＩＫ");	// IKボーン番号
            pmd.pmd_ik[n].nEffNo = Array.IndexOf(pmd_copy_bone, "左足首");		// IKターゲットボーン番号 // IKボーンが最初に接続するボーン
            pmd.pmd_ik[n].cbNumLink = 2;	// IKチェーンの長さ(子の数)
            pmd.pmd_ik[n].unCount = 40;      // 再帰演算回数 // IK値1
            pmd.pmd_ik[n].fFact = 0.5f;       // IKの影響度 // IK値2
            int[] tem_array2 = { Array.IndexOf(pmd_copy_bone, "左ひざ"), Array.IndexOf(pmd_copy_bone, "左足") }; // IK影響下のボーン番号
            pmd.pmd_ik[n++].punLinkNo = tem_array2;
 
            pmd.pmd_ik[n] = new PMD_IK();
            pmd.pmd_ik[n].nTargetNo = Array.IndexOf(pmd_copy_bone, "右つま先ＩＫ");	// IKボーン番号
            pmd.pmd_ik[n].nEffNo = Array.IndexOf(pmd_copy_bone, "右つま先");		// IKターゲットボーン番号 // IKボーンが最初に接続するボーン
            pmd.pmd_ik[n].cbNumLink = 1;	// IKチェーンの長さ(子の数)
            pmd.pmd_ik[n].unCount = 3;      // 再帰演算回数 // IK値1
            pmd.pmd_ik[n].fFact = 1;       // IKの影響度 // IK値2
            int[] tem_array3 = { Array.IndexOf(pmd_copy_bone, "右足首") }; // IK影響下のボーン番号
            pmd.pmd_ik[n++].punLinkNo = tem_array3;

            pmd.pmd_ik[n] = new PMD_IK();
            pmd.pmd_ik[n].nTargetNo = Array.IndexOf(pmd_copy_bone, "左つま先ＩＫ");	// IKボーン番号
            pmd.pmd_ik[n].nEffNo = Array.IndexOf(pmd_copy_bone, "左つま先");		// IKターゲットボーン番号 // IKボーンが最初に接続するボーン
            pmd.pmd_ik[n].cbNumLink = 1;	// IKチェーンの長さ(子の数)
            pmd.pmd_ik[n].unCount = 3;      // 再帰演算回数 // IK値1
            pmd.pmd_ik[n].fFact = 1;       // IKの影響度 // IK値2
            int[] tem_array4 = { Array.IndexOf(pmd_copy_bone, "左足首") }; // IK影響下のボーン番号
            pmd.pmd_ik[n++].punLinkNo = tem_array4;

            // -----------------------------------------------------
            // 表情情報（base以外）
            // -----------------------------------------------------
            List<Vector3[]> verPos_face = new List<Vector3[]>();

            n_face = 0; // 表情の頂点の番号（通し番号）
            foreach (TSOMesh mesh in mesh_list)
            {
                int n_inMesh = 0; // mesh中のvertexの番号

                verPos_face.Clear(); // 前回の分を消去

                for (int i = 1; i < pmd.number_of_face; i++)
                {
                    // 再読み込みして、表情を初期状態にする
                    fig.Tmo.LoadTransformationMatrixFromFrame(0);

                    // 表情に関連するboneのみ置き換えし、更新
                    foreach (String node_name in face_bone_list[pmd.pmd_face[i].cbType - 1])
                    {
                        fig.Tmo.FindNodeByName(node_name).TransformationMatrix
                            = face_tmo[i - 1].FindNodeByName(node_name).TransformationMatrix;
                    }
                    // ボーンを更新
                    fig.UpdateBoneMatricesWithoutTMOFrame();

                    // Tmo（各表情に対応する）中のBoneに従って、Mesh中の頂点位置を置き換える
                    verPos_face.Add(TsoUtils.TransVerPosWithTmo(mesh, fig.Tmo));
                }

                foreach (Vertex vertex in mesh.vertices)
                {
                    foreach (SkinWeight skin_w in vertex.skin_weights)
                    {
                        // 表情に関連するboneに影響を受ける頂点であれば、表情用の頂点とする
                        if (FACE_BONE_MIN <= mesh.bone_indices[skin_w.bone_index]
                            && mesh.bone_indices[skin_w.bone_index] <= FACE_BONE_MAX)
                        {
                            // 表情の頂点情報（base以外）
                            for (int i = 1; i < pmd.number_of_face; i++)
                            {
                                pmd.pmd_face[i].pVertices[n_face].ulIndex = n_face; // 表情用の頂点の番号(baseの番号。skin_vert_index)

                                pmd.pmd_face[i].pVertices[n_face].vec3Pos.x
                                    = verPos_face[i - 1][n_inMesh].X - pmd.pmd_face[0].pVertices[n_face].vec3Pos.x;
                                pmd.pmd_face[i].pVertices[n_face].vec3Pos.y
                                    = verPos_face[i - 1][n_inMesh].Y - pmd.pmd_face[0].pVertices[n_face].vec3Pos.y;
                                pmd.pmd_face[i].pVertices[n_face].vec3Pos.z
                                    = -verPos_face[i - 1][n_inMesh].Z - pmd.pmd_face[0].pVertices[n_face].vec3Pos.z;
                                // （bace以外は相対位置で指定）
                            }

                            n_face++;
                            break;
                        }
                    }

                    // -----------------------------------------------------
                    n_inMesh++; // mesh中のvertexの番号を一つ増やす
                }
            }


            // -----------------------------------------------------
            // 表情枠
            // -----------------------------------------------------
            pmd.skin_disp_count = pmd.number_of_face - 1; // 表情枠に表示する表情数
            pmd.skin_index = new int[pmd.skin_disp_count];
            for (int i=0; i<pmd.skin_disp_count; i++) pmd.skin_index[i] = i+1; // 表情番号

            // -----------------------------------------------------
            // ボーン枠用枠名リスト
            // -----------------------------------------------------
            pmd.bone_disp_name_count = 9; // ボーン枠用の枠名数
            pmd.disp_name = new string[pmd.bone_disp_name_count]; // 枠名(50Bytes/枠)

            pmd.disp_name[0] = "IK"         + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[1] = "体(上)"     + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[2] = "腕"         + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[3] = "指"         + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[4] = "体(下)"     + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[5] = "足"         + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[6] = "髪"         + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[7] = "乳"         + Convert.ToChar(Convert.ToInt16("0A", 16));
            pmd.disp_name[8] = "スカート"   + Convert.ToChar(Convert.ToInt16("0A", 16));
            //PMDeditorを使う場合は、枠名を0x0A00で終わらせる必要があります(0x00のみだと表示されません)。

            // -----------------------------------------------------
            // ボーン枠用表示リスト
            // -----------------------------------------------------
            List<string[]> disp_bone_list = new List<string[]>();
                    
            string[] tmp_str1 = {
                "左足ＩＫ",
                "右足ＩＫ",
                "左つま先ＩＫ",
                "右つま先ＩＫ"};
            disp_bone_list.Add(tmp_str1);
            string[] tmp_str2 = {
                "上半身", "首", "頭",
                "左目",
                "右目",
                "両目"};
            disp_bone_list.Add(tmp_str2);
            string[] tmp_str3 = {
                "左肩", "左腕", "左ひじ", "左手首",
                "右肩", "右腕", "右ひじ", "右手首"};
            disp_bone_list.Add(tmp_str3);
            string[] tmp_str4 = {
                "左親指１", "左親指２",
                "左人指１", "左人指２", "左人指３",
                "左中指１", "左中指２", "左中指３",
                "左薬指１", "左薬指２", "左薬指３",
                "左小指１", "左小指２", "左小指３",
                "右親指１", "右親指２",
                "右人指１", "右人指２", "右人指３",
                "右中指１", "右中指２", "右中指３",
                "右薬指１", "右薬指２", "右薬指３",
                "右小指１", "右小指２", "右小指３"};
            disp_bone_list.Add(tmp_str4);
            string[] tmp_str5 = {"下半身"};
            disp_bone_list.Add(tmp_str5);
            string[] tmp_str6 = {
                "左足", "左ひざ", "左足首",
                "右足", "右ひざ", "右足首"};
            disp_bone_list.Add(tmp_str6);
            string[] tmp_str7 = {
                "髪前左１", "髪前左２", "髪前左３", 
                "髪前右１", "髪前右２", "髪前右３",
                "髪横左１", "髪横左２", "髪横左３", "髪横左４", 
                "髪横右１", "髪横右２", "髪横右３", "髪横右４", 
                "髪後左１", "髪後左２", "髪後左３", "髪後左４", 
                "髪後中１", "髪後中２", "髪後中３", "髪後中４", 
                "髪後右１", "髪後右２", "髪後右３", "髪後右４"};
            disp_bone_list.Add(tmp_str7);
            string[] tmp_str8 = {
                "乳左１", "乳左２", "乳左３", "乳左４", "乳左５",
                "乳右１", "乳右２", "乳右３", "乳右４", "乳右５"};
            disp_bone_list.Add(tmp_str8);
            string[] tmp_str9 = {
                "ス前左１", "ス前左２", "ス前左３", 
                "ス前右１", "ス前右２", "ス前右３", 
                "ス横左１", "ス横左２", "ス横左３",
                "ス横右１", "ス横右２", "ス横右３",
                "ス後左１", "ス後左２", "ス後左３", 
                "ス後右１", "ス後右２", "ス後右３"};
            disp_bone_list.Add(tmp_str9);

            pmd.bone_disp_count = 
                tmp_str1.Length + tmp_str2.Length + tmp_str3.Length
                + tmp_str4.Length + tmp_str5.Length + tmp_str6.Length
                + tmp_str7.Length + tmp_str8.Length + tmp_str9.Length; // ボーン枠に表示するボーン数 (枠0(センター)を除く、すべてのボーン枠の合計)
            pmd.bone_disp = new PMD_BoneDisp[pmd.bone_disp_count]; // 枠用ボーンデータ (3Bytes/bone)

            n = 0; // 通し番号
            int n_disp = 1;
            foreach (string[] str_ in disp_bone_list)
            {
                foreach (string name in str_)
                {
                    pmd.bone_disp[n] = new PMD_BoneDisp();
                    pmd.bone_disp[n].bone_index
                        = GetBoneIDFromName(name); // 枠用ボーン番号
                    pmd.bone_disp[n++].bone_disp_frame_index
                        = n_disp; // 表示枠番号
                }
                n_disp++;
            }

            // -----------------------------------------------------
            // 英名対応(0:英名対応なし, 1:英名対応あり)
            // -----------------------------------------------------
            pmd.english_name_compatibility = 0;

            // -----------------------------------------------------
            // 剛体＆ジョイント
            // -----------------------------------------------------

            // -----------------------------------------------------
            // 身体
            body_list.Add(MakeBodyFromBone("下半身", 4));

            body_list.Add(MakeBodyFromBone("上半身", 4));
            body_list[body_list.Count - 1].shape_type = 1; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
            body_list[body_list.Count - 1].shape_w = 1.0f; // 形状：半径(幅) // CD CC CC 3F // 1.6
            body_list[body_list.Count - 1].shape_h *= 0.5f;
            body_list[body_list.Count - 1].shape_d = 0.5f; // 形状：奥行 // CD CC CC 3D // 0.1

            body_list.Add(MakeBodyFromBone("頭", 4));

            body_list.Add(MakeBodyFromBone("左肩", 4));

            body_list.Add(MakeBodyFromBone("左腕", 4));

            //body_list.Add(MakeBodyFromBone("左ひじ", 4));

            body_list.Add(MakeBodyFromBone("右肩", 4));

            body_list.Add(MakeBodyFromBone("右腕", 4));

            //body_list.Add(MakeBodyFromBone("右ひじ", 4));

            body_list.Add(MakeBodyFromBone("左足", 4));
            body_list[body_list.Count - 1].shape_w = 0.75f; // 形状：半径(幅) // CD CC CC 3F // 1.6

            body_list.Add(MakeBodyFromBone("左ひざ", 4));

            body_list.Add(MakeBodyFromBone("右足", 4));
            body_list[body_list.Count - 1].shape_w = 0.75f; // 形状：半径(幅) // CD CC CC 3F // 1.6

            body_list.Add(MakeBodyFromBone("右ひざ", 4));

            // -----------------------------------------------------
            // 身体以外
            if (kami_flag == true)
            {
                // (151, kami_Back_Mid1) - (155, kami_Back_Mid4_End)
                MakeChain("髪後中１", 0);
                body_list[GetBodyIDFromName("髪後中１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪後中２")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                joint_list[GetJointIDFromName("髪後中２-髪後中３")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後中２-髪後中３")].constrain_rot_2.x = (float)((30.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後中３-髪後中４")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後中３-髪後中４")].constrain_rot_2.x = (float)((30.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))

                // (156, kami_Front_L1) - (160, kami_Front_L4_End)
                MakeChain("髪横左１", 0);
                body_list[GetBodyIDFromName("髪横左１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪横左２")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪横左３")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従

                // (161, kami_Front_R1) - (165, kami_Front_R4_End)
                MakeChain("髪横右１", 0);
                body_list[GetBodyIDFromName("髪横右１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪横右２")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪横右３")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従

                // (166, kami_Back_L1) - (170, kami_Back_L4_End)
                MakeChain("髪後左１", 0);
                body_list[GetBodyIDFromName("髪後左１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪後左２")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                joint_list[GetJointIDFromName("髪後左２-髪後左３")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後左２-髪後左３")].constrain_rot_2.x = (float)((30.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後左３-髪後左４")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後左３-髪後左４")].constrain_rot_2.x = (float)((30.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))

                // (171, kami_Back_R1) - (175, kami_Back_R4_End)
                MakeChain("髪後右１", 0);
                body_list[GetBodyIDFromName("髪後右１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪後右２")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                joint_list[GetJointIDFromName("髪後右２-髪後右３")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後右２-髪後右３")].constrain_rot_2.x = (float)((30.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後右３-髪後右４")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("髪後右３-髪後右４")].constrain_rot_2.x = (float)((30.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))

                // (176, kami_Front_Mid1_L) - (179, kami_Front_Mid3_End_L)
                MakeChain("髪前左１", 0);
                body_list[GetBodyIDFromName("髪前左１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪前左２")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪前左３")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪前左先")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従

                // (180, kami_Front_Mid1_R) - (183, kami_Front_Mid3_End_R)
                MakeChain("髪前右１", 0);
                body_list[GetBodyIDFromName("髪前右１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪前右２")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪前右３")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                body_list[GetBodyIDFromName("髪前右先")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
            }

            if (chichi_flag == true)
            {
                // (184, Chichi_Right1) - (189, Chichi_Right5_end)
                MakeChain("乳右１", 1);
                body_list[GetBodyIDFromName("乳右１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従

                // (190, Chichi_Left1) - (195, Chichi_Left5_End)
                MakeChain("乳左１", 1);
                body_list[GetBodyIDFromName("乳左１")].rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
            }

            if (skirt_flag == true)
            {
                // (196, skirt_LeftB01) - (199, skirt_LeftB03_end)
                MakeChain("ス後左１", 2);
                joint_list[GetJointIDFromName("下半身-ス後左１")].constrain_rot_1.x = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].constrain_rot_2.x = (float)((60.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].constrain_rot_1.y = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].constrain_rot_2.y = (float)((0.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].constrain_rot_1.z = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].constrain_rot_2.z = (float)((5.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].spring_rot.x = 50.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].spring_rot.y = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後左１")].spring_rot.z = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))

                // (200, skirt_RightB01) - (203, skirt_RightB03_end)
                MakeChain("ス後右１", 2);
                joint_list[GetJointIDFromName("下半身-ス後右１")].constrain_rot_1.x = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].constrain_rot_2.x = (float)((60.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].constrain_rot_1.y = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].constrain_rot_2.y = (float)((0.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].constrain_rot_1.z = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].constrain_rot_2.z = (float)((5.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].spring_rot.x = 50.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].spring_rot.y = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス後右１")].spring_rot.z = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))

                // (204, skirt_RightS01) - (207, skirt_RightS03_end)
                MakeChain("ス横右１", 3);
                joint_list[GetJointIDFromName("下半身-ス横右１")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].constrain_rot_2.x = (float)((5.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].constrain_rot_1.y = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].constrain_rot_2.y = (float)((0.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].constrain_rot_1.z = (float)((-15.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].constrain_rot_2.z = (float)((60.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].spring_rot.x = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].spring_rot.y = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横右１")].spring_rot.z = 50.0f; // ばね：回転(rad(x), rad(y), rad(z))

                // (208, skirt_RightF01) - (211, skirt_RightF03_end)
                MakeChain("ス前右１", 2);
                joint_list[GetJointIDFromName("下半身-ス前右１")].constrain_rot_1.x = (float)((-120.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].constrain_rot_2.x = (float)((20.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].constrain_rot_1.y = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].constrain_rot_2.y = (float)((0.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].constrain_rot_1.z = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].constrain_rot_2.z = (float)((5.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].spring_rot.x = 50.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].spring_rot.y = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前右１")].spring_rot.z = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))

                // (212, skirt_LeftF01) - (215, skirt_LeftF03_end)
                MakeChain("ス前左１", 2);
                joint_list[GetJointIDFromName("下半身-ス前左１")].constrain_rot_1.x = (float)((-120.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].constrain_rot_2.x = (float)((20.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].constrain_rot_1.y = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].constrain_rot_2.y = (float)((0.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].constrain_rot_1.z = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].constrain_rot_2.z = (float)((5.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].spring_rot.x = 50.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].spring_rot.y = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス前左１")].spring_rot.z = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))

                // (216, skirt_LeftS01) - (219, skirt_LeftS03_end)
                MakeChain("ス横左１", 3);
                joint_list[GetJointIDFromName("下半身-ス横左１")].constrain_rot_1.x = (float)((-5.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].constrain_rot_2.x = (float)((5.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].constrain_rot_1.y = (float)((-0.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].constrain_rot_2.y = (float)((0.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].constrain_rot_1.z = (float)((-60.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].constrain_rot_2.z = (float)((15.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].spring_rot.x = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].spring_rot.y = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
                joint_list[GetJointIDFromName("下半身-ス横左１")].spring_rot.z = 50.0f; // ばね：回転(rad(x), rad(y), rad(z))

                // スカートの先を隣同士でジョイントする
                joint_list.Add(MakeJointFromTwoBones("ス前左２", "ス前右２"));
                joint_list.Add(MakeJointFromTwoBones("ス前右２", "ス横右２"));
                joint_list.Add(MakeJointFromTwoBones("ス横右２", "ス後右２"));
                joint_list.Add(MakeJointFromTwoBones("ス後右２", "ス後左２"));
                joint_list.Add(MakeJointFromTwoBones("ス後左２", "ス横左２"));
                joint_list.Add(MakeJointFromTwoBones("ス横左２", "ス前左２"));
                joint_list.Add(MakeJointFromTwoBones("ス前左３", "ス前右３"));
                joint_list.Add(MakeJointFromTwoBones("ス前右３", "ス横右３"));
                joint_list.Add(MakeJointFromTwoBones("ス横右３", "ス後右３"));
                joint_list.Add(MakeJointFromTwoBones("ス後右３", "ス後左３"));
                joint_list.Add(MakeJointFromTwoBones("ス後左３", "ス横左３"));
                joint_list.Add(MakeJointFromTwoBones("ス横左３", "ス前左３"));
            }

            // -----------------------------------------------------
            // 剛体＆ジョイントを配列に代入し直す
            pmd.rigidbody_count = body_list.Count;
            pmd.rigidbody = (PMD_RigidBody[])body_list.ToArray();

            pmd.joint_count = joint_list.Count;
            pmd.joint = (PMD_Joint[])joint_list.ToArray();

            // -----------------------------------------------------
            // 終了
            return "";
        }

        // ①(指定したボーン→その子ボーン)にフィットするような剛体を生成
        // ②(生成した剛体)と(指定したボーンの親ボーンの剛体)を、指定したボーンの位置でジョイントする
        // ③次に、指定したボーンの子ボーンに対しても同じ処理を行う
        // ④(①～③)を子ボーンがなくなるまで繰り返す
        public void MakeChain(string name, int flag)
        {
            if (pmd.pmd_bone[GetBoneIDFromName(name)].nChildNo != 0)
            {
                body_list.Add(MakeBodyFromBone(name, flag));
                joint_list.Add(MakeJointFromBone(name, flag));

                MakeChain(GetChildName(name), flag);
            }
            else
            {
                body_list.Add(MakeBodyFromBoneEnd(name, flag));
                joint_list.Add(MakeJointFromBone(name, flag));
            }
        }

        // ボーンの名前から、その子ボーンの名前を得る
        public string GetChildName(string name)
        {
            if (pmd.pmd_bone[GetBoneIDFromName(name)].nChildNo != 0)
            {
                return pmd.pmd_bone[pmd.pmd_bone[GetBoneIDFromName(name)].nChildNo].szName;
            }
            else
            {
                return "";
            }
        }

        // 名前からボーンIDを得る
        public int GetBoneIDFromName(string name)
        {
            for (int i = 0; i < pmd.pmd_bone.Length; i++)
            {
                if (pmd.pmd_bone[i].szName == name) return i;
            }

            return -1;
        }

        // 名前から剛体IDを得る
        public int GetBodyIDFromName(string name)
        {
            for (int i = 0; i < body_list.Count; i++)
            {
                if (body_list[i].rigidbody_name == name) return i;
            }

            return -1;
        }

        // 名前からジョイントIDを得る
        public int GetJointIDFromName(string name)
        {
            for (int i = 0; i < joint_list.Count; i++)
            {
                if (joint_list[i].joint_name == name) return i;
            }

            return -1;
        }

        // 指定したボーン位置に合わせた剛体を生成
        public PMD_RigidBody MakeBodyFromBoneEnd(string bone_name, int flag)
        {
            return (MakeBodyFromBoneEnd(GetBoneIDFromName(bone_name), flag));
        }

        // 指定したボーン位置に合わせた剛体を生成
        public PMD_RigidBody MakeBodyFromBoneEnd(int bone_num, int flag)
        {
            PMD_RigidBody rigidbody = new PMD_RigidBody();

            rigidbody.rigidbody_name = pmd.pmd_bone[bone_num].szName; // 諸データ：名称 // 頭
            rigidbody.rigidbody_rel_bone_index = bone_num; // 諸データ：関連ボーン番号 // 03 00 == 3 // 頭
            rigidbody.pos_pos.x = 0.0f;
            rigidbody.pos_pos.y = 0.0f;
            rigidbody.pos_pos.z = 0.0f;

            switch (flag)
            {
                case 0: // 髪
                    rigidbody.rigidbody_group_index = 2; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 0; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.3f; // 形状：半径(幅) // CD CC CC 3F // 1.6

                    rigidbody.rigidbody_weight = 1.0f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.8f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.8f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.8f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 1: // 乳
                    rigidbody.rigidbody_group_index = 3; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 0; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.2f; // 形状：半径(幅) // CD CC CC 3F // 1.6

                    rigidbody.rigidbody_weight = 0.01f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 2: // スカート
                    rigidbody.rigidbody_group_index = 4; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 0; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.2f; // 形状：半径(幅) // CD CC CC 3F // 1.6

                    rigidbody.rigidbody_weight = 0.5f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 3: // スカートサイド
                    rigidbody.rigidbody_group_index = 4; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 0; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.2f; // 形状：半径(幅) // CD CC CC 3F // 1.6

                    rigidbody.rigidbody_weight = 0.5f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 4: // 身体
                    rigidbody.rigidbody_group_index = 0; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = -1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 0; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.4f; // 形状：半径(幅) // CD CC CC 3F // 1.6

                    rigidbody.rigidbody_weight = 0.5f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;
            }

            return (rigidbody);
        }

        // (指定したボーン→その子ボーン)にフィットするような剛体を生成
        public PMD_RigidBody MakeBodyFromBone(string bone_name, int flag)
        {
            return (MakeBodyFromBone(GetBoneIDFromName(bone_name), flag));
        }

        // (指定したボーン→その子ボーン)にフィットするような剛体を生成
        public PMD_RigidBody MakeBodyFromBone(int bone_num, int flag)
        {
            PMD_RigidBody rigidbody = new PMD_RigidBody();

            rigidbody.rigidbody_name = pmd.pmd_bone[bone_num].szName; // 諸データ：名称 // 頭
            rigidbody.rigidbody_rel_bone_index = bone_num; // 諸データ：関連ボーン番号 // 03 00 == 3 // 頭
            
            float x1 = pmd.pmd_bone[bone_num].vec3Position.x;
            float y1 = pmd.pmd_bone[bone_num].vec3Position.y;
            float z1 = pmd.pmd_bone[bone_num].vec3Position.z;
            float x2 = pmd.pmd_bone[pmd.pmd_bone[bone_num].nChildNo].vec3Position.x;
            float y2 = pmd.pmd_bone[pmd.pmd_bone[bone_num].nChildNo].vec3Position.y;
            float z2 = pmd.pmd_bone[pmd.pmd_bone[bone_num].nChildNo].vec3Position.z;
            double L = Math.Sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1) + (z2 - z1) * (z2 - z1));
            rigidbody.pos_pos.x = (x1 + x2) / 2.0f - x1; // 位置：位置(x, y, z)
            rigidbody.pos_pos.y = (y1 + y2) / 2.0f - y1; // 位置：位置(x, y, z)
            rigidbody.pos_pos.z = (z1 + z2) / 2.0f - z1; // 位置：位置(x, y, z)
            if (y1 >= y2) rigidbody.pos_rot.x = (float)Math.Asin((z1 - z2) / L); // 位置：回転(rad(x), rad(y), rad(z))
            else rigidbody.pos_rot.x = (float)Math.Asin((z2 - z1) / L);
            if (y1 >= y2) rigidbody.pos_rot.z = (float)Math.Asin(-(x1 - x2) / L); // 位置：回転(rad(x), rad(y), rad(z))
            else rigidbody.pos_rot.z = (float)Math.Asin(-(x2 - x1) / L);

            switch (flag)
            {
                case 0: // 髪
                    rigidbody.rigidbody_group_index = 2; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 2; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.4f; // 形状：半径(幅) // CD CC CC 3F // 1.6
                    rigidbody.shape_h = (float)(L * 0.5); // 形状：高さ // CD CC CC 3D // 0.1

                    rigidbody.rigidbody_weight = 0.5f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.8f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.8f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.8f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 1: // 乳
                    rigidbody.rigidbody_group_index = 3; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 2; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.2f; // 形状：半径(幅) // CD CC CC 3F // 1.6
                    rigidbody.shape_h = (float)(L * 0.9); // 形状：高さ // CD CC CC 3D // 0.1

                    rigidbody.rigidbody_weight = 0.1f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 2: // スカート
                    rigidbody.rigidbody_group_index = 4; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 1; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.8f; // 形状：半径(幅) // CD CC CC 3F // 1.6
                    rigidbody.shape_h = (float)(L * 0.5); // 形状：高さ // CD CC CC 3D // 0.1
                    rigidbody.shape_d = 0.2f; // 形状：奥行 // CD CC CC 3D // 0.1

                    rigidbody.rigidbody_weight = 0.5f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 3: // スカートサイド
                    rigidbody.rigidbody_group_index = 4; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = 1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 1; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.2f; // 形状：半径(幅) // CD CC CC 3F // 1.6
                    rigidbody.shape_h = (float)(L * 0.5); // 形状：高さ // CD CC CC 3D // 0.1
                    rigidbody.shape_d = 0.8f; // 形状：奥行 // CD CC CC 3D // 0.1

                    rigidbody.rigidbody_weight = 0.5f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 1; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;

                case 4: // 身体
                    rigidbody.rigidbody_group_index = 0; // 諸データ：グループ // 00
                    rigidbody.rigidbody_group_target = -1; // 諸データ：グループ：対象 // 0xFFFFとの差 // 38 FE
                    rigidbody.shape_type = 2; // 形状：タイプ(0:球、1:箱、2:カプセル) // 00 // 球
                    rigidbody.shape_w = 0.4f; // 形状：半径(幅) // CD CC CC 3F // 1.6
                    rigidbody.shape_h = (float)(L * 0.8); // 形状：高さ // CD CC CC 3D // 0.1

                    rigidbody.rigidbody_weight = 0.5f; // 諸データ：質量 // 00 00 80 3F // 1.0
                    rigidbody.rigidbody_pos_dim = 0.5f; // 諸データ：移動減 // 00 00 00 00
                    rigidbody.rigidbody_rot_dim = 0.5f; // 諸データ：回転減 // 00 00 00 00
                    rigidbody.rigidbody_recoil = 0.0f; // 諸データ：反発力 // 00 00 00 00
                    rigidbody.rigidbody_friction = 0.0f; // 諸データ：摩擦力 // 00 00 00 00
                    rigidbody.rigidbody_type = 0; // 諸データ：タイプ(0:Bone追従、1:物理演算、2:物理演算(Bone位置合せ)) // 00 // Bone追従
                    break;
            }

            return (rigidbody);
        }

        // (指定したボーンにフィットさせた剛体)と、(その親ボーンにフィットさせた剛体)を、
        // 指定したボーンの位置でジョイントする
        public PMD_Joint MakeJointFromBone(string bone_name, int flag)
        {
            return (MakeJointFromBone(GetBoneIDFromName(bone_name), flag));
        }

        // (指定したボーンにフィットさせた剛体)と、(その親ボーンにフィットさせた剛体)を、
        // 指定したボーンの位置でジョイントする
        public PMD_Joint MakeJointFromBone(int bone_num, int flag)
        {
            PMD_Joint joint = new PMD_Joint();

            joint.joint_name = pmd.pmd_bone[pmd.pmd_bone[bone_num].nParentNo].szName
                + "-" + pmd.pmd_bone[bone_num].szName; // 諸データ：名称 // 右髪1
            joint.joint_rigidbody_a = GetBodyIDFromName(pmd.pmd_bone[pmd.pmd_bone[bone_num].nParentNo].szName); // 諸データ：剛体A
            joint.joint_rigidbody_b = GetBodyIDFromName(pmd.pmd_bone[bone_num].szName); // 諸データ：剛体B
            joint.joint_pos.x = pmd.pmd_bone[bone_num].vec3Position.x; // 諸データ：位置(x, y, z) // 諸データ：位置合せでも設定可
            joint.joint_pos.y = pmd.pmd_bone[bone_num].vec3Position.y; // 諸データ：位置(x, y, z) // 諸データ：位置合せでも設定可
            joint.joint_pos.z = pmd.pmd_bone[bone_num].vec3Position.z; // 諸データ：位置(x, y, z) // 諸データ：位置合せでも設定可
            joint.joint_rot.x = 0.0f; // 諸データ：回転(rad(x), rad(y), rad(z))
            joint.joint_rot.y = 0.0f; // 諸データ：回転(rad(x), rad(y), rad(z))
            joint.joint_rot.z = 0.0f; // 諸データ：回転(rad(x), rad(y), rad(z))
            joint.constrain_pos_1.x = 0.0f; // 制限：移動1(x, y, z)
            joint.constrain_pos_2.x = 0.0f; // 制限：移動2(x, y, z)
            joint.constrain_pos_1.y = 0.0f; // 制限：移動1(x, y, z)
            joint.constrain_pos_2.y = 0.0f; // 制限：移動2(x, y, z)
            joint.constrain_pos_1.z = 0.0f; // 制限：移動1(x, y, z)
            joint.constrain_pos_2.z = 0.0f; // 制限：移動2(x, y, z)
            joint.constrain_rot_1.x = 0.0f; // 制限：回転1(rad(x), rad(y), rad(z))
            joint.constrain_rot_2.x = 0.0f; // 制限：回転2(rad(x), rad(y), rad(z))
            joint.constrain_rot_1.y = 0.0f; // 制限：回転1(rad(x), rad(y), rad(z))
            joint.constrain_rot_2.y = 0.0f; // 制限：回転2(rad(x), rad(y), rad(z))
            joint.constrain_rot_1.z = 0.0f; // 制限：回転1(rad(x), rad(y), rad(z))
            joint.constrain_rot_2.z = 0.0f; // 制限：回転2(rad(x), rad(y), rad(z))
            joint.spring_pos.x = 0.0f; // ばね：移動(x, y, z)
            joint.spring_pos.y = 0.0f; // ばね：移動(x, y, z)
            joint.spring_pos.z = 0.0f; // ばね：移動(x, y, z)
            joint.spring_rot.x = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
            joint.spring_rot.y = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))
            joint.spring_rot.z = 0.0f; // ばね：回転(rad(x), rad(y), rad(z))

            switch (flag)
            {
                case 0: // 髪
                    joint.constrain_rot_1.x = (float)((-20.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                    joint.constrain_rot_2.x = (float)((20.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                    joint.constrain_rot_1.z = (float)((-20.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                    joint.constrain_rot_2.z = (float)((20.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                    joint.spring_rot.x = 10.0f; // ばね：回転(rad(x), rad(y), rad(z))
                    joint.spring_rot.y = 10.0f; // ばね：回転(rad(x), rad(y), rad(z))
                    joint.spring_rot.z = 10.0f; // ばね：回転(rad(x), rad(y), rad(z))
                    break;

                case 1: // 乳
                    joint.spring_rot.x = chichi_spring; // ばね：回転(rad(x), rad(y), rad(z))
                    joint.spring_rot.y = chichi_spring; // ばね：回転(rad(x), rad(y), rad(z))
                    joint.spring_rot.z = chichi_spring; // ばね：回転(rad(x), rad(y), rad(z))
                    break;

                case 2: // スカート
                    joint.constrain_rot_1.x = (float)((-15.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                    joint.constrain_rot_2.x = (float)((15.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                    joint.constrain_rot_1.z = (float)((-15.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
                    joint.constrain_rot_2.z = (float)((15.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
                    joint.spring_rot.x = 20.0f; // ばね：回転(rad(x), rad(y), rad(z))
                    joint.spring_rot.y = 20.0f; // ばね：回転(rad(x), rad(y), rad(z))
                    joint.spring_rot.z = 20.0f; // ばね：回転(rad(x), rad(y), rad(z))
                    break;
            }

            return (joint);
        }

        // 指定した２つのボーンにフィットさせている剛体をジョイントする
        public PMD_Joint MakeJointFromTwoBones(string bone_name1, string bone_name2)
        {
            return (MakeJointFromTwoBones(GetBoneIDFromName(bone_name1), GetBoneIDFromName(bone_name2)));
        }

        // 指定した２つのボーンにフィットさせている剛体をジョイントする
        public PMD_Joint MakeJointFromTwoBones(int bone_num1, int bone_num2)
        {
            PMD_Joint joint = new PMD_Joint();

            joint.joint_name = pmd.pmd_bone[bone_num1].szName + "-" + pmd.pmd_bone[bone_num2].szName; // 諸データ：名称 // 右髪1
            joint.joint_rigidbody_a = GetBodyIDFromName(pmd.pmd_bone[bone_num1].szName); // 諸データ：剛体A
            joint.joint_rigidbody_b = GetBodyIDFromName(pmd.pmd_bone[bone_num2].szName); // 諸データ：剛体B
            joint.joint_pos.x = 0.5f * (pmd.pmd_bone[bone_num1].vec3Position.x + pmd.pmd_bone[bone_num2].vec3Position.x); // 諸データ：位置(x, y, z) // 諸データ：位置合せでも設定可
            joint.joint_pos.y = 0.5f * (pmd.pmd_bone[bone_num1].vec3Position.y + pmd.pmd_bone[bone_num2].vec3Position.y); // 諸データ：位置(x, y, z) // 諸データ：位置合せでも設定可
            joint.joint_pos.z = 0.5f * (pmd.pmd_bone[bone_num1].vec3Position.z + pmd.pmd_bone[bone_num2].vec3Position.z); // 諸データ：位置(x, y, z) // 諸データ：位置合せでも設定可
            joint.joint_rot.x = 0.0f; // 諸データ：回転(rad(x), rad(y), rad(z))
            joint.joint_rot.y = 0.0f; // 諸データ：回転(rad(x), rad(y), rad(z))
            joint.joint_rot.z = 0.0f; // 諸データ：回転(rad(x), rad(y), rad(z))
            joint.constrain_pos_1.x = -0.5f; // 制限：移動1(x, y, z)
            joint.constrain_pos_2.x = 0.5f; // 制限：移動2(x, y, z)
            joint.constrain_pos_1.y = -0.5f; // 制限：移動1(x, y, z)
            joint.constrain_pos_2.y = 0.5f; // 制限：移動2(x, y, z)
            joint.constrain_pos_1.z = -0.5f; // 制限：移動1(x, y, z)
            joint.constrain_pos_2.z = 0.5f; // 制限：移動2(x, y, z)
            joint.constrain_rot_1.x = (float)((-60.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
            joint.constrain_rot_2.x = (float)((60.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
            joint.constrain_rot_1.y = (float)((-60.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
            joint.constrain_rot_2.y = (float)((60.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
            joint.constrain_rot_1.z = (float)((-60.0 / 180.0) * Math.PI); // 制限：回転1(rad(x), rad(y), rad(z))
            joint.constrain_rot_2.z = (float)((60.0 / 180.0) * Math.PI); // 制限：回転2(rad(x), rad(y), rad(z))
            joint.spring_pos.x = 10.0f; // ばね：移動(x, y, z)
            joint.spring_pos.y = 10.0f; // ばね：移動(x, y, z)
            joint.spring_pos.z = 10.0f; // ばね：移動(x, y, z)
            joint.spring_rot.x = 10.0f; // ばね：回転(rad(x), rad(y), rad(z))
            joint.spring_rot.y = 10.0f; // ばね：回転(rad(x), rad(y), rad(z))
            joint.spring_rot.z = 10.0f; // ばね：回転(rad(x), rad(y), rad(z))

            return (joint);
        }
    }
}
