﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using TDCG;

namespace Tso2Pmd
{
    public class TsoUtils
    {
        // Tmo中のBoneに従って、Mesh中の頂点位置及び法線ベクトルを変換して、書き出す
        // （出力する頂点情報には、頂点位置及び法線ベクトルの情報のみしか含まれていないことに注意）
        public static Vertex[] TransVerticesWithTmo(TSOMesh mesh, TMOFile tmo)
        {
            Matrix[] clipped_boneMatrices = new Matrix[mesh.maxPalettes];
            for (int numPalettes = 0; numPalettes < mesh.maxPalettes; numPalettes++)
            {
                TSONode tso_node = mesh.GetBone(numPalettes);
                TMONode tmo_node = tmo.FindNodeByName(tso_node.Name);
                clipped_boneMatrices[numPalettes] = tso_node.OffsetMatrix * tmo_node.combined_matrix;
            }

            Vertex[] output_v = new Vertex[mesh.vertices.Length];
            int n = 0;
            foreach (Vertex vertex in mesh.vertices)
            {
                Vector3 pos = Vector3.Empty;
                Vector3 nor = Vector3.Empty;

                foreach (SkinWeight sw in vertex.skin_weights)
                {
                    Matrix m = clipped_boneMatrices[sw.bone_index];

                    // 頂点位置
                    pos += Vector3.TransformCoordinate(vertex.position, m) * sw.weight;

                    // 法線ベクトル
                    m.M41 = 0;
                    m.M42 = 0;
                    m.M43 = 0;
                    nor += Vector3.TransformCoordinate(vertex.normal, m) * sw.weight;
                }

                output_v[n].position = pos;
                output_v[n++].normal = Vector3.Normalize(nor);
            }

            return output_v;
        }

        // Tmo中のBoneに従って、Mesh中の頂点位置を変換して、書き出す
        public static Vector3[] TransVerPosWithTmo(TSOMesh mesh, TMOFile tmo)
        {
            Matrix[] clipped_boneMatrices = new Matrix[mesh.maxPalettes];
            for (int numPalettes = 0; numPalettes < mesh.maxPalettes; numPalettes++)
            {
                TSONode tso_node = mesh.GetBone(numPalettes);
                TMONode tmo_node = tmo.FindNodeByName(tso_node.Name);
                clipped_boneMatrices[numPalettes] = tso_node.OffsetMatrix * tmo_node.combined_matrix;
            }

            Vector3[] output_v = new Vector3[mesh.vertices.Length];
            int n = 0;
            foreach (Vertex vertex in mesh.vertices)
            {
                Vector3 pos = Vector3.Empty;
                Vector3 nor = Vector3.Empty;

                foreach (SkinWeight sw in vertex.skin_weights)
                {
                    // 頂点位置
                    Matrix m = clipped_boneMatrices[sw.bone_index];
                    pos += Vector3.TransformCoordinate(vertex.position, m) * sw.weight;
                }

                output_v[n++] = pos;
            }

            return output_v;
        }
    }
}
