﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Diagnostics;
using System.IO;
using System.Drawing;

using TDCG;

using NyMmdUtils;
using jp.nyatla.nymmd.cs;
using jp.nyatla.nymmd.cs.struct_type;

namespace Tso2Pmd
{
    public partial class Form1 : Form
    {
        Viewer viewer;

        string[] category = {
            "身体",
            "前髪",
            "後髪",
            "頭皮(生え際)",
            "瞳",
            "ブラ",
            "全身下着・水着",
            "パンツ",
            "靴下",
            "上衣(シャツ等)",
            "全身衣装(ナース服等)",
            "上着オプション(エプロン等)",
            "下衣(スカート等)",
            "尻尾",
            "靴",
            "頭部装備(帽子等)",
            "眼鏡",
            "首輪",
            "手首",
            "背中(羽など)",
            "アホ毛類",
            "眼帯",
            "タイツ・ガーター",
            "腕装備(手甲など)",
            "リボン",
            "手持ちの小物or背景",
            "眉毛",
            "ほくろ",
            "八重歯",
            "イヤリング類"};

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            // Viewerインスタンス作成
            viewer = new Viewer();

            //ひとつのアイテムしか選択できないようにする
            listBox1.SelectionMode = SelectionMode.One;

            //ListBox1へのドロップを受け入れる
            listBox1.AllowDrop = true;
            
            //ListBox1のイベントハンドラを追加
            listBox1.DragDrop += new DragEventHandler(listBox1_DragDrop);
            
            //ListBox1にテキストを表示
            listBox1.Items.Add("ここにヘビーセーブデータをドラッグ＆ドロップするか、");
            listBox1.Items.Add("右下のボタンからセーブデータを選択してください。");
        }

        private void button1_Click(object sender, EventArgs e)
        {
            try
            {
                using (OpenFileDialog ofd = new OpenFileDialog())
                {
                    ofd.Title = "ヘビーセーブデータ(.tdcgsav.png)";
                    if (ofd.ShowDialog() != DialogResult.OK)
                    {
                        return;
                    }
                    CountVertices(ofd.FileName);
                }
            }
            catch
            {
                // エラーメッセージを表示する
                MessageBox.Show("ファイルを読み込むことができませんでした。");
            }
        }

        private void listBox1_DragEnter(object sender, DragEventArgs e)
        {
            //コントロール内にドラッグされたとき実行される
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
                //ドラッグされたデータ形式を調べ、ファイルのときはコピーとする
                e.Effect = DragDropEffects.Copy;
            else
                //ファイル以外は受け付けない
                e.Effect = DragDropEffects.None;
        }

        private void listBox1_DragDrop(object sender, DragEventArgs e)
        {
            try
            {
                //ドロップされたデータがファイルか調べる
                if (e.Data.GetDataPresent(DataFormats.FileDrop))
                {
                    string[] fileName
                        = (string[])e.Data.GetData(DataFormats.FileDrop);
                    //ドロップされたヘビーセーブを読み込む
                    CountVertices(fileName[0]);
                }
            }
            catch
            {
                // エラーメッセージを表示する
                MessageBox.Show("ファイルを読み込むことができませんでした。");
            }
        }

        private void CountVertices(string file_name)
        {
            // ヘビーセーブファイルを開く
            viewer.ClearFigureList(); // 前回読み込んだフィギュアを消去
            List<Figure> fig_list = viewer.LoadPNGFile(file_name);

            // カテゴリを得る
            List<string> opt = new List<string>();
            PNGFile png = new PNGFile();
            png.Ftso += delegate(Stream dest, int extract_length, byte[] opt1)
            {
                opt.Add(category[(int)opt1[0]]);
            };
            png.Load(file_name);

            // リストボックスを空に
            listBox1.Items.Clear();

            // 頂点数をチェック
            int verCount = 0;
            int n_tso = 0;
            foreach (TSOFile tso in fig_list[0].TSOList)
            {
                listBox1.Items.Add("---------------------------------");
                listBox1.Items.Add(" " + opt[n_tso++]);
                listBox1.Items.Add("---------------------------------");

                int verCount_inTso = 0;

                foreach (TSOFrame frame in tso.frames)
                {
                    int verCount_inFrame = 0;

                    foreach (TSOMesh mesh in frame.meshes)
                    {
                        verCount += mesh.vertices.Length;
                        verCount_inFrame += mesh.vertices.Length;
                        verCount_inTso += mesh.vertices.Length;
                    }

                    listBox1.Items.Add(frame.name + " : " + verCount_inFrame.ToString());
                }

                listBox1.Items.Add("");
                listBox1.Items.Add("TSO中での合計頂点数 " + " : " + verCount_inTso.ToString());
                listBox1.Items.Add("---------------------------------");
                listBox1.Items.Add("");
                listBox1.Items.Add("");
            }
            label_VerCount.Text = verCount.ToString();

            if (verCount < 65535)
            {
                label_VerCount.ForeColor = Color.Black;
            }
            else
            {
                label_VerCount.ForeColor = Color.Red;
            }
        }
    }
}
